"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

import numpy as np
import sympy as sp
from numba import njit

from qrisp import fast_append
from qrisp.simulator.bi_arrays import BiArray, DenseBiArray, SparseBiArray, tensordot


np_dtype = np.complex64

id_matrix = np.eye(2, dtype=np_dtype)

pauli_x = np.asarray([[0, 1], [1, 0]], dtype=np_dtype)
pauli_y = np.asarray([[0, 0], [0, 0]]) + 1j * np.asarray([[0, -1], [1, 0]], dtype = np_dtype)
pauli_z = np.asarray([[1, 0], [0, -1]], dtype=np_dtype)


# Function which returns the unitary of a u3 gate
def u3matrix(theta, phi, lam, global_phase):
    import numpy as module

    res = module.zeros(shape=(2, 2), dtype=module.complex64)

    from sympy.core.expr import Expr

    I = 1j
    for par in [theta, phi, lam, global_phase]:
        if isinstance(par, Expr):
            res = module.zeros(shape=(2, 2), dtype="object")
            import sympy as module

            I = sp.I
            break

    res[0, 0] = module.cos(theta / 2)
    res[0, 1] = -module.exp(I * lam) * module.sin(theta / 2)
    res[1, 0] = module.exp(I * phi) * module.sin(theta / 2)
    res[1, 1] = module.exp(I * (phi + lam)) * module.cos(theta / 2)

    return res * module.exp(I * global_phase)


# Efficient function to generate the unitary of a controlled gate
def controlled_unitary(controlled_gate):
    m = controlled_gate.base_operation.num_qubits
    try:
        temp = controlled_gate.base_operation.unitary_array
    except AttributeError:
        temp = (
            controlled_gate.base_operation.unitary_array
        ) = controlled_gate.base_operation.get_unitary()

    n = controlled_gate.num_qubits
    res = np.eye(2**n, dtype=temp.dtype)

    res[-(2**m) :, -(2**m) :] = temp

    controlled_gate.unitary = temp

    return res


# Calculates the unitary of a gate embeddded into a circuit of n qubits
# For instance consider this circuit
# Here we have gate = CXGate(), n = 4, destination_qubits = [2,3]
# q80_0: ─────


# q80_1: ─────
#        ┌───┐
# q80_2: ┤ X ├
#        └─┬─┘
# q80_3: ──■──
# This function then returns the overall unitary matrix in the form of either
# a BiArray or a numpy array depending what is more efficient
def calc_embedded_unitary(gate, n, destination_qubits):
    # First we generate the array corresponding to
    # 1 (x) 1 (x) ... gate
    # where 1 denotes the identity mapping on a qubit,
    # (x) the tensor product,
    # and gate the gate unitary

    # In principle, we could use the numpy function np.kron
    # tensor_array = np.kron(np.eye(2**(n-m)), gate.get_unitary())

    # However, this performs alot of unnecessary multiplications since the structure of
    # this matrix is simply the unitary of the gate on the diagonal 2**n/2**dim(gate)
    # times repeated
    if gate.definition is not None:
        gate_array = __calc_circuit_unitary(gate.definition)
        gate.unitary = gate_array
    else:
        gate_array = gate.get_unitary()

    # Numpy arrays are only fast on a low scale than BiArrays. Therefore,  we generate
    # a Numpy Arrays
    if n < 10 and not gate_array.dtype == np.dtype("O"):
        tensor_array = generate_id_kron_jitted(gate_array, n)
    else:
        tensor_array = generate_id_kron(gate_array, n)

    # Reshape into an 2n-dimensional array where each axis represents an
    # index of the tensor factors
    tensor_array = tensor_array.reshape(n * [2, 2])

    # #Generate the nec<essary swaps to put the destination qubits at the right place
    swaps = generate_swaps(destination_qubits, n)

    # Perform swaps
    for swap in swaps:
        tensor_array = swap_tensor_factors(tensor_array, *swap)

    # Shape back into the 2**n x 2**n matrix
    tensor_array = tensor_array.reshape([1 << n, 1 << n])

    if tensor_array.dtype == np.dtype("O") and n < 10:
        tensor_array = tensor_array.to_array()

    return tensor_array


# In order to generate the swaps we follow the strategy that
# each destination qubit can be guaranteed to be puth in the correct place
# with a single swap. Therefore, we iterate through the destination qubits and
# successively put them in the right place
def generate_swaps(destination_qubits, n):
    # This list represents the current state of the tensor
    # The list element represents an identity tensor
    tensor_state = (n - len(destination_qubits)) * ["id"] + list(destination_qubits)
    swaps = []

    # Function to check if the tensor state reached its desired state
    def check_if_reached(tensor_state):
        for i in range(len(tensor_state)):
            if tensor_state[i] != "id":
                if tensor_state[i] != i:
                    return False
        return True

    # Repat until the tensor reached its final state
    while not check_if_reached(tensor_state):
        # Iterate through the tensor state and find a qubit which is not
        # in the correct position.
        for i in range(len(tensor_state)):
            if tensor_state[i] != "id":
                # If tensor_state[i] == i the qubit in question already has reached the
                # final state
                if tensor_state[i] != i:
                    # Swap the list elements
                    index = int(i)
                    value = int(tensor_state[i])
                    tensor_state[index], tensor_state[value] = (
                        tensor_state[value],
                        tensor_state[index],
                    )

                    # Log the swaps
                    swaps.append((index, value))

                    # Restart search for qubit in an incorrect position
                    break

    return list(swaps)


# Swaps the axis of a kroneker product (tensor product of linear mappings)
# ie. if we insert tensor = 1 (x) A (x) 1 (x) 1
# and apply this function with i = 0, j = 1
# we get A (x) 1 (x) 1 (x) 1
def swap_tensor_factors(tensor, i, j):
    n = len(tensor.shape) // 2

    # Note that we have to swap two axes because we are
    # dealing with the kroneker product and not the tensor product

    tensor = tensor.swapaxes(i, j)
    tensor = tensor.swapaxes(n + i, n + j)

    return tensor


# In principle we could use the numpy function np.kron
# tensor_array = np.kron(np.eye(2**(n-m)), gate.get_unitary())


# However this performs alot of unnecessary multiplications
# since the structure of this matrix is simply the unitary
# of the gate on the diagonal 2**n/2**dim(gate) times repeated
def generate_id_kron(input_tensor, n):
    if isinstance(input_tensor, np.ndarray):
        if input_tensor.dtype == np.dtype("O"):
            input_tensor = DenseBiArray(input_tensor)
        else:
            input_tensor = SparseBiArray(input_tensor)

    k = int(np.log2(input_tensor.shape[0]))

    data = np.ones(2 ** (n - k))

    nz_indices = np.arange(2 ** (n - k)).astype(np.int64)
    nz_indices = nz_indices * 2 ** (n - k) + nz_indices

    id_tensor = SparseBiArray((nz_indices, data), shape=(2 ** (n - k), 2 ** (n - k)))

    id_tensor.reshape((n - k) * [2, 2])
    input_tensor.reshape(k * [2, 2])

    res = tensordot(id_tensor, input_tensor, ((), ()))

    for i in range(n - k):
        for j in range(k):
            res.swapaxes(2 * n - 2 * k - 1 + j - i, 2 * n - 2 * k + j - i)

    res.reshape([2**n, 2**n])

    return res


@njit(cache=True)
def generate_id_kron_jitted(input_tensor, n):
    res = np.zeros((2**n, 2**n), dtype=input_tensor.dtype)

    m = int(input_tensor.shape[0])
    for i in range(2**n // m):
        res[i * m : (i + 1) * m, i * m : (i + 1) * m] = input_tensor

    return res


# This functions calculates the unitary of a given circuit

# The idea to make this more efficient is that not all matrix multiplication have to be
# perforned in the 2**n dimensional basis

# For instance in this circuit:

#      ┌───┐
# q_0: ┤ X ├──■──
#      └───┘┌─┴─┐
# q_1: ─────┤ X ├
#           └───┘
# q_2: ──────────

# q_3: ──────────

# We can calculate the unitary on the first two qubits on the 4 d basis
# and then embedd this unitary into the 2**4 = 16d basis


# In order to harness the effciency gain we use a divide and conquer strategy:
# We merge the instruction of the circuit into pairs of elementary gates, calculate
# this unitary and the again merge this unitary with the neighbouring pair of elementary
# gate. Applying this recursively requires only a single 2**n d matrix multiplication.


def __calc_circuit_unitary(qc):
    n = len(qc.qubits)

    # If the circuit is empty, return the identity matrix
    if len(qc.data) == 0:
        return np.eye(2**n, dtype = np_dtype)

    # If the circuit contains only a single insturction,
    # calculate this instructions unitary and embedd it
    if len(qc.data) == 1:
        instr_0 = qc.data[0]
        unitary_0 = calc_embedded_unitary(
            qc.data[0].op, n, [qc.qubits.index(qb) for qb in instr_0.qubits]
        )

        return unitary_0

    # If the circuit is a pair of instructions and uses all qubits,
    # we calculate their product by matrix multiplication

    # If the circuit contained unused qubits, the next "else" statement will turn
    # this circuit into it's version without any idle qubits
    # and this case is then called at a later recursion
    elif len(qc.data) == 2 and len(qc.qubits) == len(
        set(qc.data[0].qubits).union(qc.data[1].qubits)
    ):
        instr_0 = qc.data[0]
        instr_1 = qc.data[1]

        unitary_0 = calc_embedded_unitary(
            instr_0.op, n, [qc.qubits.index(qb) for qb in instr_0.qubits]
        )
        unitary_1 = calc_embedded_unitary(
            instr_1.op, n, [qc.qubits.index(qb) for qb in instr_1.qubits]
        )

        res = tensordot(unitary_1, unitary_0, (1, 0), contract_sparsity_threshold=0.01)

    # This is the case where we generate the pairs of instructions
    else:
        # Set up a new circuit where the merged instructions are appended
        recursion_qc = qc.clearcopy()

        # Merge every pair of neighbouring instructions into one instruction
        for i in range(0, len(qc.data), 2):
            if i + 1 < len(qc.data):
                recursion_qc.append(qc.data[i].merge(qc.data[i + 1]))
            else:
                recursion_qc.append(qc.data[i])
        # The recursion circuit now has about half of the instructions of the original
        # quantum circuit. Therefore, the recursion is finite
        res = __calc_circuit_unitary(recursion_qc)

    return res


# The result of __calc_circuit_unitary can either be a numpy array or a BiArray.
# This is important in order for the recursion structure to work out
# This function is a wrapper that converts to the desired output
def calc_circuit_unitary(qc, res_type="numpy"):
    with fast_append():
        res = __calc_circuit_unitary(qc)

    if res_type == "numpy":
        if isinstance(res, BiArray):
            res = res.to_array()

        return res

    elif res_type == "bi_array":
        if isinstance(res, BiArray):
            return res

        return DenseBiArray(res)

    else:
        raise Exception("Don't know result type " + res_type)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 27.04.2022
import pytest

from qrisp.logic_synthesis import TruthTable, gray_logic_synth
from qrisp.misc import int_encoder
from qrisp.core import QuantumVariable, QuantumSession


@pytest.mark.parametrize("num", ["001", "011", "000", "100"])
def test_gray_synthesis_example(num):
    # split the input string into a list of characters
    tmp = list(num)

    # Create a testing truth table
    tt = TruthTable(tmp)

    # Create new quantum session
    qs = QuantumSession()

    # Create input variable
    input_var = QuantumVariable(tt.bit_amount, qs)

    # Encode index

    int_encoder(input_var, 0)

    # Create output variable
    output_var = QuantumVariable(tt.shape[1], qs)

    # Perform logic synthesis
    gray_logic_synth(input_var, output_var, tt, phase_tolerant=False)

    # Expected bitstring comes from tt in format array [0 0 1]
    # create string for assert comparison
    ttstr = "".join([str(item) for item in tt.n_rep[0, :]])
    mes_res = output_var.get_measurement()

    assert ttstr == str(list(mes_res.keys())[0])
    assert ttstr in ["001", "011", "000", "100"] and str(list(mes_res.keys())[0]) in [
        "001",
        "011",
        "000",
        "100",
    ]



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.environments.quantum_environments import *
from qrisp.environments.quantum_conditionals import *
from qrisp.environments.quantum_inversion import *
from qrisp.environments.control_environment import *
from qrisp.environments.gate_wrap_environment import *
# from qrisp.environments.temp_var_environment import *
from qrisp.environments.conjugation_environment import *
from qrisp.environments.GMS_environment import *
from qrisp.environments.iteration_environment import *




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np
from qrisp.circuit import Operation


class GXX_wrapper(Operation):
    def __init__(self, n, chi):
        if n == 1:
            pass
            # raise Exception("Tried to create 1D GXX-gate")
        flattened_chi = sum(chi, [])

        super().__init__(name="GXX", num_qubits=n, params=flattened_chi)

        self.chi = chi
        self.n = n
        from qrisp.circuit import QuantumCircuit

        qc = QuantumCircuit(n)

        for i in range(len(chi)):
            for j in range(len(chi[i])):
                if chi[i][j] == 0 or i == j:
                    continue
                qc.rxx(chi[i][j] / 2, i, j)

        self.definition = qc

    def inverse(self):
        chi_inv = [[-x for x in y] for y in self.chi]
        return GXX_wrapper(self.n, chi_inv)


# Mimics a circuit which performs n cphase gates, where each cp interacts with the last
# qubit.
# example
# q2874_0: ─■─────────────────────────────────────────────────────────────
#           │
# q2874_1: ─┼────────■────────────────────────────────────────────────────
#           │        │
# q2874_2: ─┼────────┼────────■───────────────────────────────────────────
#           │        │        │
# q2874_3: ─┼────────┼────────┼────────■──────────────────────────────────
#           │        │        │        │
# q2874_4: ─┼────────┼────────┼────────┼────────■─────────────────────────
#           │        │        │        │        │
# q2874_5: ─┼────────┼────────┼────────┼────────┼────────■────────────────
#           │        │        │        │        │        │
# q2874_6: ─┼────────┼────────┼────────┼────────┼────────┼────────■───────
#           │P(π/2)  │P(π/2)  │P(π/2)  │P(π/2)  │P(π/2)  │P(π/2)  │P(π/2)
# q2874_7: ─■────────■────────■────────■────────■────────■────────■───────
# Using only one GMS gate or two uniform GMS gates
def gms_multi_cp_gate_mono_phase(
    n, theta, use_uniform=True, phase_tolerant=False, basis="GXX"
):
    from qrisp import QuantumSession, QuantumVariable, cp, h, p

    qs = QuantumSession()
    qv = QuantumVariable(n + 1, qs)
    if basis == "GXX":
        if use_uniform:
            gms_1 = GXX_wrapper((n + 1), (n + 1) * [(n + 1) * [-theta / 2]])
            gms_2 = GXX_wrapper(n, n * [n * [theta / 2]])
        else:
            # gms = GXX_wrapper(n, (n+1)*[n*[0] + [-theta/2]])
            for i in range(n):
                cp(theta, qv[i], qv[n])
                # qs.cp(theta, i, n)
            return GXX_converter(qs).to_gate()

        if not phase_tolerant:
            for i in range(n):
                p(theta / 2, qv[i])
                # qs.p(theta/2, qv.reg[i])

        p(theta / 2 * n, qv[-1])
        # qs.p(theta/2*n, qv.reg[-1])

        h(qv)

        qs.append(gms_1, qv.reg)
        if not phase_tolerant:
            qs.append(gms_2, qv.reg[:-1])

        h(qv)

    elif basis == "GZZ":
        if use_uniform:
            gms_1 = GZZ_wrapper((n + 1), (n + 1) * [(n + 1) * [-theta / 2]])
            gms_2 = GZZ_wrapper(n, n * [n * [theta / 2]])
        else:
            # gms = GXX_wrapper(n, (n+1)*[n*[0] + [-theta/2]])
            for i in range(n):
                cp(theta, qv[i], qv[n])
                # qs.cp(theta, i, n)
            return GZZ_converter(qs).to_gate()

        if not phase_tolerant:
            for i in range(n):
                p(theta / 2, qv[i])
                # qs.p(theta/2, qv.reg[i])

        p(theta / 2 * n, qv[-1])
        # qs.p(theta/2*n, qv.reg[-1])

        qs.append(gms_1, qv.reg)
        if not phase_tolerant:
            qs.append(gms_2, qv.reg[:-1])

    else:
        raise "Basis choice must be GXX or GZZ"

    return qs.to_gate(name="gms_mono_phase_cp")


# qb_0: ──■─────────────────
#         │
# qb_1: ──┼────■────────────
#         │    │
# qb_2: ──┼────┼────■───────
#         │    │    │
# qb_3: ──┼────┼────┼────■──
#       ┌─┴─┐┌─┴─┐┌─┴─┐┌─┴─┐
# qb_4: ┤ X ├┤ X ├┤ X ├┤ X ├
#       └───┘└───┘└───┘└───┘
def gms_multi_cx_fan_in(n, use_uniform=True, phase_tolerant=False, basis="GXX"):
    from qrisp import QuantumSession, QuantumVariable, h

    qs = QuantumSession()
    qv = QuantumVariable(n + 1, qs)

    h(qv[-1])

    qs.append(
        gms_multi_cp_gate_mono_phase(
            n,
            np.pi,
            use_uniform=use_uniform,
            phase_tolerant=phase_tolerant,
            basis=basis,
        ),
        qv.reg,
    )

    h(qv[-1])

    # result = qs.to_gate(name = "gms_multi_cx_fan_in")
    # result.name = "gms_multi_cx_fan_in"

    return qs.to_gate(name="gms_multi_cx_fan_in")


# qubit_0: ──■────■────■────■──
#          ┌─┴─┐  │    │    │
# qubit_1: ┤ x ├──┼────┼────┼──
#          └───┘┌─┴─┐  │    │
# qubit_2: ─────┤ x ├──┼────┼──
#               └───┘┌─┴─┐  │
# qubit_3: ──────────┤ x ├──┼──
#                    └───┘┌─┴─┐
# qubit_4: ───────────────┤ x ├
#                         └───┘
def gms_multi_cx_fan_out(n, use_uniform=True, phase_tolerant=False, basis="GXX"):
    from qrisp import QuantumSession, QuantumVariable, h

    qs = QuantumSession()
    qv = QuantumVariable(n + 1, qs)

    for i in range(n):
        h(qv[i])

    qs.append(
        gms_multi_cp_gate_mono_phase(
            n,
            np.pi,
            use_uniform=use_uniform,
            phase_tolerant=phase_tolerant,
            basis=basis,
        ),
        qv.reg,
    )

    for i in range(n):
        h(qv[i])

    result = qs.to_gate()
    result.name = "gms_multi_cx_fan_out"

    return result


# This functions takes a circuit which only consists of phase and cphase gates
# and turns it into a GXX gate + some one-qubit gates
# The basic idea is that a cphase gate acts as
# CP |ab> = exp(i theta/2 (1 - (-1)^(ab))) |ab>
# We rewrite (-1)^(ab) = -1/2*((-1)^(a+b) - ((-1)^a + (-1)^b) - 1)
# The term (-1)^(a+b) can now be executed by a Moelmer-Soerensen:
# exp(i theta/2 Z_a (x) Z_b)
# The terms (-1)^a and (-1)^b are executed by single qubit phase gates
# The remaining -1 is an irrelevant global phase


def GXX_converter(qs):
    from qrisp import QuantumSession, QuantumVariable

    # Check if the given Circuit is valid
    global_phase = np.array([0.0])
    for i in range(len(qs.data)):
        if not qs.data[i].op.name in [
            "p",
            "cp",
            "id",
            "cz",
            "rz",
            "qb_alloc",
            "qb_dealloc",
        ]:
            raise Exception(qs.data[i].op.name + " is neither Phase nor CPhase gate")

        if not qs.data[i].op.name in ["cp", "cz", "qb_alloc", "qb_dealloc"]:
            global_phase += qs.data[i].op.global_phase

        # Subtract a 1/4 of the phase in case we are dealing with a cp gate (for more
        # elaboration, check eq. 99 of https://arxiv.org/abs/2112.10537)
        if qs.data[i].op.name == "cp":
            global_phase -= qs.data[i].op.params[0] / 4

        elif qs.data[i].op.name == "cz":
            global_phase -= np.pi / 4

    # Create qubit index dictionary
    qubit_index_dic = {qs.qubits[i]: i for i in range(len(qs.qubits))}

    # Create phase matrix
    # The (i,j) entry of this matrix contains the phase that is applied onto the
    # qubit pair (i,j), the indices with i == j represent single qubit rz gates
    n = len(qs.qubits)
    phase_matrix = np.zeros((n, n))

    # Go through all data entries of the circuit
    for i in range(len(qs.data)):
        # Set instruction alias
        ins = qs.data[i]

        if ins.op.name in ["id", "qb_alloc", "qb_dealloc"]:
            continue

        # If sinlge qubit gate => Collect phase
        if len(ins.qubits) == 1:
            index = qubit_index_dic[ins.qubits[0]]

            phase_matrix[index, index] += ins.op.params[0]
            continue

        # Get indices
        index_0 = qubit_index_dic[ins.qubits[0]]
        index_1 = qubit_index_dic[ins.qubits[1]]

        # Swap indices if neccessary such that the phase matrix is an upper triangle
        # matrix
        if index_0 >= index_1:
            temp = int(index_0)
            index_0 = index_1
            index_1 = temp

        # Collect phase
        if ins.op.name == "cz":
            phase_matrix[index_0, index_1] += np.pi
        else:
            phase_matrix[index_0, index_1] += ins.op.params[0]

    # Create Quantum Session
    qc_res = qs.clearcopy()

    if global_phase[0] != 0:
        qc_res.gphase(global_phase[0], 0)

    # Calculate the row-sum and the column sum of the phase matrix
    # to determine which single qubit phase has to be applied
    # This is because every entry with the same row / column represents a phase gate
    # where the qubit in question participated
    for i in range(n):
        qc_res.p((sum(phase_matrix[i, :]) + sum(phase_matrix[:, i])) / 2, i)
        phase_matrix[i, i] = 0

    # Prepary Chi list for GXX gate
    chi_list = [[-phase_matrix[j, i] for i in range(n)] for j in range(n)]

    # Apply GXX Gate
    qc_res.h(qc_res.qubits)
    qc_res.append(GXX_wrapper(n, chi_list), range(n))
    qc_res.h(qc_res.qubits)

    return qc_res


# Similar to gms_multi_cp_gate_mono_phase but also allows more than one phase
# ie. converts circuits of the type
# q54264_0: ─■─────────────────────────────────────────
#            │
# q54264_1: ─┼─────────■───────────────────────────────
#            │         │
# q54264_2: ─┼─────────┼──────────■────────────────────
#            │         │          │
# q54264_3: ─┼─────────┼──────────┼──────────■─────────
#            │P(-π/5)  │P(-2π/5)  │P(-3π/5)  │P(-4π/5)
# q54264_4: ─■─────────■──────────■──────────■─────────
# The advantage over the GXX_converter is the use_uniform mode, which makes
# sure only uniform GMS gates are used
def gms_multi_cp_gate(n, phases, use_uniform=True, basis="GXX"):
    from qrisp import QuantumSession, QuantumVariable, p

    qs = QuantumSession()

    qv = QuantumVariable(n + 1, qs)

    for i in range(n):
        p(phases[i] / 2, qv[i])
        # qs.p(phases[i]/2, i)

    p(sum(phases) / 2, qv[n])
    # qs.p(sum(phases)/2, n)

    qs.append(
        gms_multi_cx_fan_out(
            n, use_uniform=use_uniform, phase_tolerant=False, basis=basis
        ),
        qv.reg,
    )

    for i in range(n):
        qs.p(-phases[i] / 2, i)

    qs.append(
        gms_multi_cx_fan_out(
            n, use_uniform=use_uniform, phase_tolerant=False, basis=basis
        ).inverse(),
        qv.reg,
    )

    result = qs.to_gate()

    result.name = "gms_multi_cp_gate"

    return result


class GZZ_wrapper(Operation):
    def __init__(self, n, chi):
        if n == 1:
            pass
            # raise Exception("Tried to create 1D GZZ-gate")
        flattened_chi = sum(chi, [])

        super().__init__(name="GZZ", num_qubits=n, params=flattened_chi)

        self.chi = chi
        self.n = n
        from qrisp.circuit import QuantumCircuit

        qc = QuantumCircuit(n)

        for i in range(len(chi)):
            for j in range(len(chi[i])):
                if chi[i][j] == 0 or i == j:
                    continue
                qc.rzz(chi[i][j] / 2, i, j)

        self.definition = qc


# This functions takes a circuit which only consists of phase and cphase gates
# and turns it into a GZZ gate + some one-qubit gates
# The basic idea is that a cphase gate acts as
# CP |ab> = exp(i theta/2 (1 - (-1)^(ab))) |ab>
# We rewrite (-1)^(ab) = -1/2*((-1)^(a+b) - ((-1)^a + (-1)^b) - 1)
# The term (-1)^(a+b) can now be executed by a Moelmer-Soerensen:
# exp(i theta/2 Z_a (x) Z_b)
# The terms (-1)^a and (-1)^b are executed by single qubit phase gates
# The remaining -1 is an irrelevant global phase
def GZZ_converter(qs):
    from qrisp import QuantumSession, QuantumVariable

    # Check if the given Circuit is valid
    global_phase = np.array([0.0])
    for i in range(len(qs.data)):
        if not qs.data[i].op.name in [
            "p",
            "cp",
            "id",
            "cz",
            "rz",
            "qb_alloc",
            "qb_dealloc",
        ]:
            raise Exception(qs.data[i].op.name + " is neither Phase nor CPhase gate")

        if not qs.data[i].op.name in ["cp", "cz", "qb_alloc", "qb_dealloc"]:
            global_phase += qs.data[i].op.global_phase

        # Subtract a 1/4 of the phase in case we are dealing with a cp gate (for more
        # elaboration, check eq. 99 of https://arxiv.org/abs/2112.10537)
        if qs.data[i].op.name == "cp":
            global_phase -= qs.data[i].op.params[0] / 4

        elif qs.data[i].op.name == "cz":
            global_phase -= np.pi / 4

    # Create qubit index dictionary
    qubit_index_dic = {qs.qubits[i]: i for i in range(len(qs.qubits))}

    # Create phase matrix
    # The (i,j) entry of this matrix contains the phase that is applied onto the
    # qubit pair (i,j), the indices with i == j represent single qubit rz gates
    n = len(qs.qubits)
    phase_matrix = np.zeros((n, n))

    # Go through all data entries of the circuit
    for i in range(len(qs.data)):
        # Set instruction alias
        ins = qs.data[i]

        if ins.op.name in ["id", "qb_alloc", "qb_dealloc"]:
            continue

        # If sinlge qubit gate => Collect phase
        if len(ins.qubits) == 1:
            index = qubit_index_dic[ins.qubits[0]]

            phase_matrix[index, index] += ins.op.params[0]
            continue

        # Get indices
        index_0 = qubit_index_dic[ins.qubits[0]]
        index_1 = qubit_index_dic[ins.qubits[1]]

        # Swap indices if neccessary such that the phase matrix is an upper triangle
        # matrix
        if index_0 >= index_1:
            temp = int(index_0)
            index_0 = index_1
            index_1 = temp

        # Collect phase
        if ins.op.name == "cz":
            phase_matrix[index_0, index_1] += np.pi
        else:
            phase_matrix[index_0, index_1] += ins.op.params[0]

    # Create Quantum Session
    qc_res = qs.clearcopy()

    if global_phase[0] != 0:
        qc_res.gphase(global_phase[0], 0)

    # Calculate the row-sum and the column sum of the phase matrix
    # to determine which single qubit phase has to be applied
    # This is because every entry with the same row / column represents a phase gate
    # where the qubit in question participated
    for i in range(n):
        qc_res.p((sum(phase_matrix[i, :]) + sum(phase_matrix[:, i])) / 2, i)
        phase_matrix[i, i] = 0

    # Prepary Chi list for GZZ gate
    chi_list = [[-phase_matrix[j, i] for i in range(n)] for j in range(n)]

    # Apply GZZ Gate
    qc_res.append(GZZ_wrapper(n, chi_list), range(n))

    return qc_res



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

#
# Autogenerated by Thrift Compiler (0.15.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

import sys

from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import (
    TApplicationException,
    TException,
    TFrozenDict,
    TMessageType,
    TType,
)
from thrift.transport import TTransport
from thrift.TRecursive import fix_spec

all_structs = []


class Qubit(object):
    """
    Attributes:
     - identifier

    """

    def __init__(
        self,
        identifier=None,
    ):
        self.identifier = identifier

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.identifier = (
                        iprot.readString().decode("utf-8", errors="replace")
                        if sys.version_info[0] == 2
                        else iprot.readString()
                    )
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("Qubit")
        if self.identifier is not None:
            oprot.writeFieldBegin("identifier", TType.STRING, 1)
            oprot.writeString(
                self.identifier.encode("utf-8")
                if sys.version_info[0] == 2
                else self.identifier
            )
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Clbit(object):
    """
    Attributes:
     - identifier

    """

    def __init__(
        self,
        identifier=None,
    ):
        self.identifier = identifier

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.identifier = (
                        iprot.readString().decode("utf-8", errors="replace")
                        if sys.version_info[0] == 2
                        else iprot.readString()
                    )
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("Clbit")
        if self.identifier is not None:
            oprot.writeFieldBegin("identifier", TType.STRING, 1)
            oprot.writeString(
                self.identifier.encode("utf-8")
                if sys.version_info[0] == 2
                else self.identifier
            )
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Operation(object):
    """
    Attributes:
     - name
     - num_qubits
     - num_clbits
     - params
     - definition

    """

    def __init__(
        self,
        name=None,
        num_qubits=None,
        num_clbits=None,
        params=None,
        definition=None,
    ):
        self.name = name
        self.num_qubits = num_qubits
        self.num_clbits = num_clbits
        self.params = params
        self.definition = definition

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.name = (
                        iprot.readString().decode("utf-8", errors="replace")
                        if sys.version_info[0] == 2
                        else iprot.readString()
                    )
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.num_qubits = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.num_clbits = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.params = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = iprot.readDouble()
                        self.params.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRUCT:
                    self.definition = QuantumCircuit()
                    self.definition.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("Operation")
        if self.name is not None:
            oprot.writeFieldBegin("name", TType.STRING, 1)
            oprot.writeString(
                self.name.encode("utf-8") if sys.version_info[0] == 2 else self.name
            )
            oprot.writeFieldEnd()
        if self.num_qubits is not None:
            oprot.writeFieldBegin("num_qubits", TType.I32, 2)
            oprot.writeI32(self.num_qubits)
            oprot.writeFieldEnd()
        if self.num_clbits is not None:
            oprot.writeFieldBegin("num_clbits", TType.I32, 3)
            oprot.writeI32(self.num_clbits)
            oprot.writeFieldEnd()
        if self.params is not None:
            oprot.writeFieldBegin("params", TType.LIST, 4)
            oprot.writeListBegin(TType.DOUBLE, len(self.params))
            for iter6 in self.params:
                oprot.writeDouble(iter6)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.definition is not None:
            oprot.writeFieldBegin("definition", TType.STRUCT, 5)
            self.definition.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Instruction(object):
    """
    Attributes:
     - op
     - qubits
     - clbits

    """

    def __init__(
        self,
        op=None,
        qubits=None,
        clbits=None,
    ):
        self.op = op
        self.qubits = qubits
        self.clbits = clbits

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.op = Operation()
                    self.op.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.qubits = []
                    (_etype10, _size7) = iprot.readListBegin()
                    for _i11 in range(_size7):
                        _elem12 = Qubit()
                        _elem12.read(iprot)
                        self.qubits.append(_elem12)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.clbits = []
                    (_etype16, _size13) = iprot.readListBegin()
                    for _i17 in range(_size13):
                        _elem18 = Clbit()
                        _elem18.read(iprot)
                        self.clbits.append(_elem18)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("Instruction")
        if self.op is not None:
            oprot.writeFieldBegin("op", TType.STRUCT, 1)
            self.op.write(oprot)
            oprot.writeFieldEnd()
        if self.qubits is not None:
            oprot.writeFieldBegin("qubits", TType.LIST, 2)
            oprot.writeListBegin(TType.STRUCT, len(self.qubits))
            for iter19 in self.qubits:
                iter19.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.clbits is not None:
            oprot.writeFieldBegin("clbits", TType.LIST, 3)
            oprot.writeListBegin(TType.STRUCT, len(self.clbits))
            for iter20 in self.clbits:
                iter20.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class QuantumCircuit(object):
    """
    Attributes:
     - qubits
     - clbits
     - data
     - init

    """

    def __init__(
        self,
        qubits=None,
        clbits=None,
        data=None,
        init=True,
    ):
        self.qubits = qubits
        self.clbits = clbits
        self.data = data
        self.init = init

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.qubits = []
                    (_etype24, _size21) = iprot.readListBegin()
                    for _i25 in range(_size21):
                        _elem26 = Qubit()
                        _elem26.read(iprot)
                        self.qubits.append(_elem26)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.clbits = []
                    (_etype30, _size27) = iprot.readListBegin()
                    for _i31 in range(_size27):
                        _elem32 = Clbit()
                        _elem32.read(iprot)
                        self.clbits.append(_elem32)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.data = []
                    (_etype36, _size33) = iprot.readListBegin()
                    for _i37 in range(_size33):
                        _elem38 = Instruction()
                        _elem38.read(iprot)
                        self.data.append(_elem38)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.BOOL:
                    self.init = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("QuantumCircuit")
        if self.qubits is not None:
            oprot.writeFieldBegin("qubits", TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.qubits))
            for iter39 in self.qubits:
                iter39.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.clbits is not None:
            oprot.writeFieldBegin("clbits", TType.LIST, 2)
            oprot.writeListBegin(TType.STRUCT, len(self.clbits))
            for iter40 in self.clbits:
                iter40.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.data is not None:
            oprot.writeFieldBegin("data", TType.LIST, 3)
            oprot.writeListBegin(TType.STRUCT, len(self.data))
            for iter41 in self.data:
                iter41.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.init is not None:
            oprot.writeFieldBegin("init", TType.BOOL, 4)
            oprot.writeBool(self.init)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Message(object):
    """
    Attributes:
     - qubits
     - annotation

    """

    def __init__(
        self,
        qubits=None,
        annotation=None,
    ):
        self.qubits = qubits
        self.annotation = annotation

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.qubits = []
                    (_etype45, _size42) = iprot.readListBegin()
                    for _i46 in range(_size42):
                        _elem47 = Qubit()
                        _elem47.read(iprot)
                        self.qubits.append(_elem47)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.annotation = (
                        iprot.readString().decode("utf-8", errors="replace")
                        if sys.version_info[0] == 2
                        else iprot.readString()
                    )
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("Message")
        if self.qubits is not None:
            oprot.writeFieldBegin("qubits", TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.qubits))
            for iter48 in self.qubits:
                iter48.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.annotation is not None:
            oprot.writeFieldBegin("annotation", TType.STRING, 2)
            oprot.writeString(
                self.annotation.encode("utf-8")
                if sys.version_info[0] == 2
                else self.annotation
            )
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


all_structs.append(Qubit)
Qubit.thrift_spec = (
    None,  # 0
    (
        1,
        TType.STRING,
        "identifier",
        "UTF8",
        None,
    ),  # 1
)
all_structs.append(Clbit)
Clbit.thrift_spec = (
    None,  # 0
    (
        1,
        TType.STRING,
        "identifier",
        "UTF8",
        None,
    ),  # 1
)
all_structs.append(Operation)
Operation.thrift_spec = (
    None,  # 0
    (
        1,
        TType.STRING,
        "name",
        "UTF8",
        None,
    ),  # 1
    (
        2,
        TType.I32,
        "num_qubits",
        None,
        None,
    ),  # 2
    (
        3,
        TType.I32,
        "num_clbits",
        None,
        None,
    ),  # 3
    (
        4,
        TType.LIST,
        "params",
        (TType.DOUBLE, None, False),
        None,
    ),  # 4
    (
        5,
        TType.STRUCT,
        "definition",
        [QuantumCircuit, None],
        None,
    ),  # 5
)
all_structs.append(Instruction)
Instruction.thrift_spec = (
    None,  # 0
    (
        1,
        TType.STRUCT,
        "op",
        [Operation, None],
        None,
    ),  # 1
    (
        2,
        TType.LIST,
        "qubits",
        (TType.STRUCT, [Qubit, None], False),
        None,
    ),  # 2
    (
        3,
        TType.LIST,
        "clbits",
        (TType.STRUCT, [Clbit, None], False),
        None,
    ),  # 3
)
all_structs.append(QuantumCircuit)
QuantumCircuit.thrift_spec = (
    None,  # 0
    (
        1,
        TType.LIST,
        "qubits",
        (TType.STRUCT, [Qubit, None], False),
        None,
    ),  # 1
    (
        2,
        TType.LIST,
        "clbits",
        (TType.STRUCT, [Clbit, None], False),
        None,
    ),  # 2
    (
        3,
        TType.LIST,
        "data",
        (TType.STRUCT, [Instruction, None], False),
        None,
    ),  # 3
    (
        4,
        TType.BOOL,
        "init",
        None,
        True,
    ),  # 4
)
all_structs.append(Message)
Message.thrift_spec = (
    None,  # 0
    (
        1,
        TType.LIST,
        "qubits",
        (TType.STRUCT, [Qubit, None], False),
        None,
    ),  # 1
    (
        2,
        TType.STRING,
        "annotation",
        "UTF8",
        None,
    ),  # 2
)
fix_spec(all_structs)
del all_structs



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by nik40643 at 23.06.2023

import time
import random
import numpy as np

from qrisp.core import QuantumSession, QuantumVariable
from qrisp import rz, rx, ry, p, cp
import qiskit.circuit.library.standard_gates as qsk_gates
from qiskit.circuit import Parameter, QuantumCircuit, ParameterExpression
#from qrisp.circuit.quantum_circuit import rxx, ryy, rzz
from sympy import Symbol, simplify, symbols
from qrisp.interface import circuit_converter, convert_from_qiskit

def test_abstract_params_printing():
    qiskit_qc = QuantumCircuit(5)
    
    params = symbols("c:5")
    #gates_list = [qsk_gates.RXGate(),qsk_gates.RYGate(),qsk_gates.RZGate(),qsk_gates.PhaseGate(),qsk_gates.CPhaseGate()]
    qiskit_ins_list = []
    theta = Parameter(str(params[0]))
    qiskit_qc.append(qsk_gates.RXGate(theta ), [0],[])
    qiskit_qc.append(qsk_gates.RYGate(Parameter(str(params[1]))), [1],[])
    qiskit_qc.append(qsk_gates.RZGate(Parameter(str(params[2]))), [2],[])
    qiskit_qc.append(qsk_gates.PhaseGate(Parameter(str(params[3]))), [3],[])
    #qiskit_qc.append(qsk_gates.CPhaseGate(Parameter(str(params[4]))), [4],[])
    #print(str(qiskit_qc.qasm()))
    str(qiskit_qc)
    #print(qiskit_qc)
    
    
    qv = QuantumVariable(5)
    rx(params[0], qv[0])
    ry(params[1], qv[1])
    rz(params[2], qv[2])
    p(params[3], qv[3])
    #print(str(qv.qs))
    #print(qv.qs)
    qisk = circuit_converter.convert_to_qiskit(qv.qs)
    #print(qisk)
    
    #assert str(qisk) == str(qiskit_qc)
    qisk_list = []
    for gate in qisk.data:
        qisk_list.append((gate[0].name,gate[0].params))
    
    qiskit_list = []
    for gate in qiskit_qc.data:
        qiskit_list.append((gate[0].name,gate[0].params))
    
    assert len(qisk_list) == len(qisk_list)
    for index in range(len(qisk_list)):
        assert str(qiskit_list[index]) == str(qisk_list[index])
    
    
    ### After conversion from Qiskit
    
    theto = Symbol("theta")
    gummo = Symbol("gamma")
    #theto = Parameter("theta")
    #gummo = Parameter("gamma")
    #delto = theto +gummo
    delto = 2*Parameter("theta") 
    
    n = 5
    
    qc = QuantumCircuit(5, 1)
    qc.h(0)
    for i in range(n-1):
        qc.cx(i, i+1)
    qc.rz(delto, range(5))
    #qc.barrier()
    for i in reversed(range(n-1)):
        qc.cx(i, i+1)
    qc.h(0)
    qc.measure(0, 0)
    
    qrisp_qc = convert_from_qiskit(qc)
    print(qrisp_qc)
    
    
    
    assert len(list((qrisp_qc.abstract_params))) == len(list(qc.parameters ))   
    
    delto = 2*Parameter("theta") + Parameter("gamma")
    n = 5
    qc = QuantumCircuit(5, 1)
    qc.h(0)
    for i in range(n-1):
        qc.cx(i, i+1)
    qc.rx(delto, range(5))
    #qc.barrier()
    for i in reversed(range(n-1)):
        qc.cx(i, i+1)
    qc.h(0)
    qc.measure(0, 0)
    
    qrisp_qc = convert_from_qiskit(qc)
    
    
    assert len(list((qrisp_qc.abstract_params))) == len(list(qc.parameters ))   
    
    
    # write some test to assert lambd_expr = ParameterExpression.sympify(p) 


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from itertools import product

import numpy as np

from qrisp.circuit import transpile
from qrisp.core import QuantumVariable
from qrisp.misc import bin_rep
# from qrisp.environments.temp_var_environment import temp_qv
from qrisp.environments import conjugate

class QuantumArray(np.ndarray):
    """
    This class allows the convenient management of multiple QuantumVariables of one
    type. As a subclass of `numpy's ndarray
    <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, the
    QuantumArray supports many convenient array manipulation methods. Similar to the
    numpy equivalent, creating a QuantumArray can be achieved by specifying a shape and
    a ``qtype``:

    >>> import numpy as np
    >>> from qrisp import QuantumArray, QuantumFloat
    >>> qtype = QuantumFloat(5, -2)
    >>> q_array = QuantumArray(qtype = qtype, shape = (2, 2, 2))

    Note that ``qtype`` is not a type object but a QuantumVariable which serves as an
    "example".

    To retrieve the entries (i.e. QuantumVariables) from the QuantumArray, we simply
    index as with regular numpy arrays:

    >>> from qrisp import h
    >>> qv = q_array[0,0,1]
    >>> h(qv[0])
    >>> print(q_array)
    {OutcomeArray([[[0., 0.],
                    [0., 0.]],
                   [[0., 0.],
                    [0., 0.]]]): 0.5,
     OutcomeArray([[[0.  , 0.25],
                    [0.  , 0.  ]],
                   [[0.  , 0.  ],
                    [0.  , 0.  ]]]): 0.5}

    We see the value 0.25 in the second entry because we applied an H-gate onto the 0-th
    qubit of the QuantumVariable at position (0,0,1). Since the type of this array is a
    QuantumFloat, with exponent -2, the significance of this qubit is 0.25.

    Note that the keys of the dictionary returned by the get_measurement method are no
    regular numpy arrays, as key objects need to be hashable. Instead, they are objects
    of an immutable subclass of np.ndarray called OutcomeArray, that supports hashing.

    For QuantumArrays, many methods known from numpy arrays work here too:

    >>> q_array = q_array.reshape(2,4)

    Not only do the ndarray methods work but also many other convenience functions from
    the numpy module:

    >>> q_array_swap = np.swapaxes(q_array, 0, 1)
    >>> print(q_array_swap)
    {OutcomeArray([[0., 0.],
                   [0., 0.],
                   [0., 0.],
                   [0., 0.]]): 0.5,
     OutcomeArray([[0.  , 0.  ],
                   [0.25, 0.  ],
                   [0.  , 0.  ],
                   [0.  , 0.  ]]): 0.5}

    To initiate the array, we use the :meth:`encode <qrisp.QuantumArray.encode>` method.
    Similar to QuantumVariables, we can also use the slicing operator, but this time
    non-trivial slices are possible as well:

    >>> q_array[1:,:] = 2*np.ones((1,4))
    >>> print(q_array)
    {OutcomeArray([[0., 0., 0., 0.],
                   [2., 2., 2., 2.]]): 0.5,
     OutcomeArray([[0.  , 0.25, 0.  , 0.  ],
                   [2.  , 2.  , 2.  , 2.  ]]): 0.5}


    The shape of a QuantumArray does not have to be specified at creation. We can either
    set it through the :meth:`set_shape <qrisp.QuantumArray.set_shape>` method or by
    initiating:

    >>> q_array_1 = QuantumArray(qtype = qtype)
    >>> q_array_1.set_shape((2,2))
    >>> print(q_array_1)
    {OutcomeArray([[0., 0.],
                   [0., 0.]]): 1.0}
    >>> q_array_2 = QuantumArray(qtype = qtype)
    >>> q_array_2[:] = np.eye(2)
    >>> print(q_array_2)
    {OutcomeArray([[1., 0.],
                   [0., 1.]]): 1.0}

    **Quantum indexing**

    QuantumArrays can be dereferenced by :ref:`QuantumFloats <QuantumFloat>`. This
    returns a :ref:`QuantumEnvironment` in which the corresponding entry is avaliable as
    a QuantumVariable. ::

        from qrisp import QuantumBool, QuantumArray, QuantumFloat, h, multi_measurement

        q_array = QuantumArray(QuantumBool(), shape = (4,4))
        index_0 = QuantumFloat(2)
        index_1 = QuantumFloat(2)


        index_0[:] = 2
        index_1[:] = 1

        h(index_0[0])

        with q_array[index_0, index_1] as entry:
            entry.flip()

    >>> print(multi_measurement([index_0, index_1, q_array]))
    {(2, 1, OutcomeArray([[0., 0., 0., 0.],
                          [0., 0., 0., 0.],
                          [0., 1., 0., 0.],
                          [0., 0., 0., 0.]])): 0.5,
     (3, 1, OutcomeArray([[0., 0., 0., 0.],
                          [0., 0., 0., 0.],
                          [0., 0., 0., 0.],
                          [0., 1., 0., 0.]])): 0.5}

    .. note::
        This only works for arrays which have a size of an integer power of 2.

    **Matrix multiplication**

    For QuantumArrays with ``qtype`` QuantumFloat, matrix multiplication is available.

    >>> q_array_1 = QuantumArray(qtype)
    >>> q_array_2 = QuantumArray(qtype)
    >>> q_array_1[:] = 2*np.eye(2)
    >>> q_array_2[:] = [[1,2],[3,4]]
    >>> print(q_array_1 @ q_array_2)
    {OutcomeArray([[2., 4.],
                   [6., 0.]]): 1.0}

    .. note::
        By default, the output matrix will have the same ``qtype`` as the first input
        matrix. Here, the ``qtype`` is a QuantumFloat with 5 mantissa bits and exponent
        -2, implying that the result 8 yields overflow. Since qrisps unsigend arithmetic
        is modular, we get a 0.

    It is also possible to multiply classical and quantum matrices

    >>> q_array = QuantumArray(qtype)
    >>> q_array[:] = 3*np.eye(2)
    >>> cl_array = np.array([[1,2],[3,4]])
    >>> print(q_array @ cl_array)
    {OutcomeArray([[3., 6.],
                   [1., 4.]]): 1.0}
    """

    def __new__(subtype, qtype, shape=0, qs=None, name=None):
        if isinstance(shape, QuantumVariable):
            raise Exception

        obj = super().__new__(subtype, shape, dtype="object")

        if shape == 0:
            obj.shape_specified = False
        else:
            obj.shape_specified = True

        from qrisp.misc import find_calling_line

        if name is None:
            if type(obj) is QuantumArray:
                line = find_calling_line(1)
            else:
                line = find_calling_line(2)

            split_line = line.split("=")

            if len(split_line) == 2 and False:
                python_var_name = split_line[0]
                python_var_name = python_var_name.split(" ")[0]
                python_var_name = python_var_name.split(" ")[-1]
                name = python_var_name
            else:
                from qrisp import QuantumBool, QuantumChar, QuantumFloat

                if isinstance(qtype, QuantumFloat):
                    name = qtype.get_unique_name("qf_array")
                elif isinstance(qtype, QuantumBool):
                    name = qtype.get_unique_name("qbl_array")
                elif isinstance(qtype, QuantumChar):
                    name = qtype.get_unique_name("qch_array")
                else:
                    name = qtype.get_unique_name("qv_array")

        else:
            if name[-1] == "*":
                name = qtype.get_unique_name(name[:-1])

        obj.name = qtype.name

        # Set data size
        obj.msize = qtype.size
        obj.qtype = qtype

        if qs is None:
            from qrisp.core import QuantumSession

            qs = QuantumSession()

        if isinstance(shape, int):
            shape = (shape,)
        indices = list(product(*[list(range(i)) for i in shape]))

        for i in indices:
            
            temp_dup = qtype.duplicate(name = obj.name + "*", qs=qs)
            np.ndarray.__setitem__(obj, i, temp_dup)
            
        obj.qs = qs

        return obj

    def decoder(self, code_int):
        """
        The decoder method specifies how a QuantumArray turns the outcomes of
        measurements into human-readable values. It recieves an integer i and returns an
        OutcomeArray.

        Parameters
        ----------
        i : int
            Integer representing the outcome of a measurement of the qubits of this
            QuantumArray.

        Returns
        -------
        res : np.ndarray
            An array with entries of the type of the results of the .decoder of the
            qtype of this array.

        Examples
        --------

        We create a QuantumFloat and inspect its decoder:

        >>> from qrisp import QuantumArray, QuantumFloat
        >>> qtype = QuantumFloat(3)
        >>> q_array = QuantumArray(qtype, (2,2))
        >>> print(q_array.decoder(1))
        [[0 0]
         [0 1]]

        """

        flattened_array = self.flatten()

        from qrisp.qtypes.quantum_float import QuantumFloat

        if isinstance(self.qtype, QuantumFloat):
            if self.qtype.exponent >= 0:
                res = np.zeros(len(flattened_array), dtype=np.int64)
            else:
                res = np.zeros(len(flattened_array))
        else:
            res = np.zeros(len(flattened_array))

        n = len(self.qtype)

        bin_string = bin_rep(code_int, len(flattened_array) * n)

        for i in range(len(flattened_array)):
            try:
                res[i] = self.qtype.decoder(int(bin_string[i * n : (i + 1) * n], 2))
            except ValueError:
                res = res.astype("object")
                res[i] = self.qtype.decoder(int(bin_string[i * n : (i + 1) * n], 2))

        return OutcomeArray(res.reshape(self.shape))

    def encoder(self, encoding_array):
        """
        The encoder reverses the decoder, it turns arrays into integers based on the
        encoder of the ``qtype`` of this array.

        Parameters
        ----------
        encoding_array :
            An array where the entries can be read by the decoder of qtype.

        Raises
        ------
        Exception
            Tried to call encoder on array with mismatching shape.

        Returns
        -------
        i : int
            The integer encoding the given array.

        Examples
        --------

        We create a QuantumArray and inspect it's encoder:

        >>> from qrisp import QuantumArray, QuantumFloat
        >>> qtype = QuantumFloat(3)
        >>> q_array = QuantumArray(qtype, (2,2))
        >>> print(q_array.encoder(np.eye(2)))
        513
        """

        if isinstance(encoding_array, list):
            encoding_array = np.array(encoding_array)

        if self.shape != encoding_array.shape:
            raise Exception("Tried to call encoder on array with mismatching shape")

        flattened_encoding_array = encoding_array.flatten()
        flattened_quantum_array = self.flatten()

        encoding_int_str = ""

        for i in range(len(flattened_encoding_array)):
            qv = flattened_quantum_array[i]
            encoding_int_str += bin_rep(
                qv.encoder(flattened_encoding_array[i]), qv.size
            )[::-1]

        return int(encoding_int_str[::-1], 2)

    def set_shape(self, shape):
        """
        Method to specify a shape for arrays which have been created without one.

        Parameters
        ----------
        shape : tuple
            Desired shape.

        Raises
        ------
        Exception
            QuantumArray already has a shape.

        """

        if self.shape_specified:
            raise Exception(
                "Tried to set shape of QuantumArray, which already has a shape"
            )

        if isinstance(shape, int):
            shape = (shape,)

        self.resize(shape, refcheck=False)

        indices = list(product(*[list(range(i)) for i in shape]))

        for i in indices:
            temp_dup = self.qtype.duplicate(qs=self.qs)
            np.ndarray.__setitem__(self, i, temp_dup)

        self.shape_specified = True

    def __setitem__(self, key, value):
        if not self.shape_specified:
            if not hasattr(value, "shape"):
                value = np.array(value)
            self.set_shape(value.shape)

        if isinstance(value, QuantumArray):
            self[key].init_from(value)
            return

        if isinstance(value, QuantumVariable):
            return
        self[key].encode(value)
        return self

    def __array_function__(self, func, types, args, kwargs):
        if func.__name__ == "concatenate":
            from qrisp import merge

            if id(args[0][0].qtype) != id(args[0][1].qtype):
                raise Exception(
                    "Tried to concatenate arrays with differing qtype objects"
                )

            merge(args)

            ndarray_res = np.ndarray.__array_function__(self, func, types, args, kwargs)

            res = QuantumArray(self.qtype, shape=ndarray_res.shape)

            indices = product(*[list(range(i)) for i in ndarray_res.shape])
            for i in indices:
                np.ndarray.__setitem__(res, i, ndarray_res[i])

            res.qs = args[0][0].qs
            res.qtype = args[0][0].qtype

            return res

        else:
            return np.ndarray.__array_function__(self, func, types, args, kwargs)

    # Retrieves a measurement of the arrays state
    # Returns a list of tuples of the type (array, count)
    # ie. [(array([1,1,0]), 232), (array([1,1,3]), 115), ...]
    def get_measurement(
        self,
        backend=None,
        shots=10000,
        compile=True,
        compilation_kwargs={},
        subs_dic={},
        circuit_preprocessor=None,
    ):
        """
        Method for acquiring measurement results for the given array. The semantics are
        similar to the :meth:`get_measurement <qrisp.QuantumVariable.get_measurement>`
        method of QuantumVariable. The results are returned as a dictionary of another
        numpy subtype called OutcomeArray.

        Parameters
        ----------
        backend : BackendClient, optional
            The backend on which to evaluate the quantum circuit. The default can be
            specified in the file default_backend.py.
        shots : integer, optional
            The amount of shots to evaluate the circuit. The default is 10000.
        compile : bool, optional
            Boolean indicating if the .compile method of the underlying QuantumSession
            should be called before. The default is True.
        compilation_kwargs  : dict, optional
            Keyword arguments for the compile method. For more details check
            :meth:`QuantumSession.compile <qrisp.QuantumSession.compile>`.
            The default is ``{}``.
        subs_dic : dict, optional
            A dictionary of sympy symbols and floats to specify parameters in the case
            of a circuit with unspecified, abstract parameters. The default is {}.
        circuit_preprocessor : Python function, optional
            A function which recieves a QuantumCircuit and returns one, which is applied
            after compilation and parameter substitution. The default is None.

        Raises
        ------
        Exception
            Tried to get measurement within open environment.

        Returns
        -------
        list of tuples
            The measurement results in the form [(outcome_label, probability), ...].

        Examples
        --------

        >>> from qrisp import QuantumFloat, QuantumArray
        >>> qtype = QuantumFloat(3)
        >>> q_array = QuantumArray(qtype)
        >>> q_array[:] = [[1,0],[0,1]]
        >>> res = q_array.get_measurement()
        >>> print(res)
        {OutcomeArray([[1, 0],
                       [0, 1]]): 1.0}
        """

        for qv in self.flatten():
            if qv.is_deleted():
                raise Exception(
                    "Tried to measure QuantumArray containing deleted QuantumVariables"
                )

        if not self.shape_specified:
            raise Exception("Tried to measure QuantumArray without shape specification")

        if backend is None:
            if self.qs.backend is None:
                from qrisp.default_backend import def_backend

                backend = def_backend
            else:
                backend = self.qs.backend

        if len(self.qs.env_stack) != 0:
            raise Exception("Tried to get measurement within open environment")

        qubits = sum([qv.reg for qv in self.flatten()[::-1]], [])
        # Copy circuit in over to prevent modification
        from qrisp.quantum_network import QuantumNetworkClient

        if isinstance(backend, QuantumNetworkClient):
            qc = self.qs.copy()
            self.qs.clear_data()
            shots = 1
        else:
            if compile:
                from qrisp.core.compilation import qompiler

                qc = qompiler(
                    self.qs, intended_measurements=qubits, **compilation_kwargs
                )
                # qc = self.qs.compile(compilation_workspace)
            else:
                qc = self.qs.copy()

        # Bind parameters
        if subs_dic:
            qc = qc.bind_parameters(subs_dic)

        # Execute user specified circuit_preprocessor
        if circuit_preprocessor is not None:
            qc = circuit_preprocessor(qc)

        # Transpile circuit
        qc = transpile(qc)

        from qrisp.misc import get_measurement_from_qc

        counts = get_measurement_from_qc(qc, qubits, backend, shots)

        # Insert outcome labels (if available and hashable)
        new_counts_dic = {}
        for key in counts.keys():
            outcome_label = self.decoder(key)

            new_counts_dic[outcome_label] = counts[key]

        counts = new_counts_dic

        # Sort keys
        sorted_key_list = list(counts.keys())
        sorted_key_list.sort(key=lambda x: -counts[x])
        counts = {key: counts[key] for key in sorted_key_list}

        return counts

    def encode(self, encoding_array):
        """
        The encode method allows to quickly bring a QuantumArray in a desired
        computational basis state. For this, it performs a circuit, bringing fresh
        qubits into the integer state specified by the encoder.

        A shorthand for this method is given by the ``[:]`` operator.

        Note that the qubits to initialize have to be fresh (i.e. no operations
        performed on them).

        Parameters
        ----------
        value :
            A value supported by the encoder.

        Examples
        --------

        We create a QuantumArray and encode the identity matrix.

        >>> from qrisp import QuantumArray, QuantumFloat
        >>> qtype = QuantumFloat(5)
        >>> q_array = QuantumArray(qtype, (4,4))
        >>> q_array.encode(np.eye(4))
        >>> print(q_array)
        {OutcomeArray([[1, 0, 0, 0],
                       [0, 1, 0, 0],
                       [0, 0, 1, 0],
                       [0, 0, 0, 1]]): 1.0}

        Using the slice operator we can also encode on slices of QuantumArrays

        >>> q_array = QuantumArray(qtype, (4,4))
        >>> q_array[:,:2] = np.ones((4,2))
        >>> print(q_array)
        {OutcomeArray([[1, 1, 0, 0],
                       [1, 1, 0, 0],
                       [1, 1, 0, 0],
                       [1, 1, 0, 0]]): 1.0}
        """

        flattened_array = self.flatten()

        qubit_list = sum([qv.reg for qv in flattened_array], [])

        from qrisp.misc import check_if_fresh, int_encoder

        if not check_if_fresh(qubit_list, self.qs):
            raise Exception("Tried to initialize qubits which are not fresh anymore.")

        int_encoder(qubit_list, self.encoder(encoding_array))

        return

    def delete(self, verify=False):
        r"""
        Performs the :meth:`delete <qrisp.QuantumVariable.delete>` method on all
        QuantumVariables in this array.

        Parameters
        ----------
        verify : bool, optional
            If this keyword is set to true, a simulator is queried to check if the
            deleted qubits are in the $\ket{0}$ state. The default is False.

        """
        temp = self.flatten()
        for i in range(len(temp)):
            temp.qv_array[i].delete()

    def __repr__(self):
        return str(self.get_measurement())

    def __str__(self):
        return str(self.get_measurement())

    def __matmul__(self, other):
        from qrisp import QuantumFloat

        if isinstance(self.qtype, QuantumFloat):
            if isinstance(other, QuantumArray):
                from qrisp.arithmetic import q_matmul

                return q_matmul(self, other)

            elif isinstance(other, np.ndarray):
                from qrisp.arithmetic import semi_classic_matmul

                return semi_classic_matmul(self, other)

        raise Exception("Matrix multiplication for non-float types not implemented")

    def __rmatmul__(self, other):
        from qrisp import QuantumFloat

        if isinstance(self.qtype, QuantumFloat):
            return (self.transpose() @ other.transpose()).transpose()

    def __getitem__(self, index):
        if isinstance(index, QuantumVariable):
            return manipulate_array(self, index)

        if isinstance(index, tuple):
            if isinstance(index[0], QuantumVariable):
                return manipulate_array(self, index)

        return np.ndarray.__getitem__(self, index)

    def init_state(self, tuple_list):
        """
        Method to initiate arbitrary quantum states in this array. The semantics are
        similar to the :meth:`QuantumVariable equivalent
        <qrisp.QuantumVariable.init_state>` of this method.

        The given state will be normalized.

        Parameters
        ----------
        tuple_list : list of tuples
            The list of tuples describing the quantum state. The first componenet of the
            tuples needs to represent the array and the second the required amplitude.

        Raises
        ------
        Exception
            Tried to initialize quantum state on qubits which are not fresh anymore.

        Examples
        --------

        We initiate a quantum state on an array and evaluate the measurement
        probabilities.

        >>> from qrisp import QuantumArray, QuantumFloat
        >>> qtype = QuantumFloat(3)
        >>> q_array = QuantumArray(qtype, shape = 3)
        >>> q_array.init_state([(np.array([1, 2, 3]), 1), (np.array([1, 2, 2]), 0.5j)])
        >>> print(q_array)
        {OutcomeArray([1, 2, 3]): 0.8, OutcomeArray([1, 2, 2]): 0.2}

        """

        if isinstance(tuple_list, dict):
            tuple_list = [(k, v) for k, v in tuple_list.items()]

        flattened_array = self.flatten()
        qubit_list = []

        for qv in flattened_array:
            qubit_list += qv.reg
            from qrisp.misc import check_if_fresh

            if not check_if_fresh(qv.reg, qv.qs):
                raise Exception(
                    "Tried to initialize quantum state on qubits,"
                    "which are not fresh anymore"
                )

        target_array = np.zeros(2 ** len(qubit_list), dtype=np.complex128)

        for i in range(len(tuple_list)):
            target_array[self.encoder(tuple_list[i][0])] = tuple_list[i][1]

        target_array = target_array / np.vdot(target_array, target_array) ** 0.5

        from qrisp import init_state

        init_state(qubit_list, target_array)

    def init_from(self, other):
        """
        Performs the :meth:`init_from <qrisp.QuantumVariable.init_from>` method on all
        containing QuantumVariables based on their equivalent in the QuantumArray other.
        Note that this method does not copy the quantum state. For more details check
        the documentation of the init_from method of QuantumVariable.

        A shorthand for this method is the slicing operator ``[:]``.

        Parameters
        ----------
        other : QuantumArray
            The QuantumArray from which to initiate.

        Raises
        ------
        Exception
            Tried to initialize from array of invalied shape or qtype.

        Examples
        --------

        We create a QuantumArray and bring it into superposition

        >>> import numpy as np
        >>> from qrisp import QuantumArray, QuantumFloat, h, multi_measurement
        >>> qtype = QuantumFloat(5)
        >>> q_array_a = QuantumArray(qtype)
        >>> q_array_a[:] = np.eye(3)
        >>> h(q_array_a[0,0][0])
        >>> print(q_array_a)
        {OutcomeArray([[0, 0, 0],
                       [0, 1, 0],
                       [0, 0, 1]]): 0.5,
         OutcomeArray([[1, 0, 0],
                       [0, 1, 0],
                       [0, 0, 1]]): 0.5}

        We now duplicate this array and initiate

        >>> q_array_b = q_array_a.duplicate()
        >>> q_array_b.init_from(q_array_a)
        >>> print(multi_measurement([q_array_a, q_array_b]))
                {(OutcomeArray([[0, 0, 0],
                                [0, 1, 0],
                                [0, 0, 1]]),
                  OutcomeArray([[0, 0, 0],
                                [0, 1, 0],
                                [0, 0, 1]])): 0.5,
                 (OutcomeArray([[1, 0, 0],
                                [0, 1, 0],
                                [0, 0, 1]]),
                  OutcomeArray([[1, 0, 0],
                                [0, 1, 0],
                                [0, 0, 1]])): 0.5}


        We can achieve the same with the slicing operator

        >>> q_array_c = QuantumArray(qtype, shape = (3, 3))
        >>> q_array_c[1,:] = q_array_a[0,:]
        >>> print(multi_measurement([q_array_b, q_array_c]))
        {(OutcomeArray([[0, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0]]),
          OutcomeArray([[0, 0, 0],
                        [0, 0, 0],
                        [0, 0, 0]])): 0.5,
         (OutcomeArray([[0, 0, 0],
                        [1, 0, 0],
                        [0, 0, 0]]),
          OutcomeArray([[0, 0, 0],
                        [1, 0, 0],
                        [0, 0, 0]])): 0.5}
        """
        if id(self.qtype) != id(other.qtype) or self.shape != other.shape:
            raise Exception("Tried to initialize from array of invalid shape or qtype")

        self_fl = self.flatten()
        other_fl = other.flatten()

        for i in range(len(self_fl)):
            self_fl[i].init_from(other_fl[i])

    def duplicate(self, init=False, qs=None):
        """
        This method returns a fresh QuantumArray, with equal ``qtype`` and shape.

        Parameters
        ----------
        init : bool, optional
            If set to True, the :meth:`init_from <qrisp.QuantumArray.init_from>` method
            will be called after creation. The default is False.
        qs : QuantumSession, optional
            The QuantumSession where the duplicate should be registered. By default,
            the duplicate will be registered in a new QuantumSession.

        Returns
        -------
        res : QuantumArray
            The duplicated array.

        Examples
        --------

        We duplicate a QuantumArray consisting of QuantumFloats with and without
        initiation.


        >>> from qrisp import QuantumArray, QuantumFloat
        >>> qtype = QuantumFloat(4)
        >>> q_array_0 = QuantumArray(qtype, (2,2))
        >>> q_array_0[:] = np.ones((2,2))
        >>> print(q_array_0)
        {OutcomeArray([[1, 1],
                       [1, 1]]): 1.0}
        >>> q_array_1 = q_array_0.duplicate()
        >>> print(q_array_1)
        {OutcomeArray([[0, 0],
                       [0, 0]]): 1.0}

        Note that no values have been carried over:

        >>> q_array_2 = q_array_0.duplicate(init = True)
        >>> print(q_array_2)
        {OutcomeArray([[1, 1],
                       [1, 1]]): 1.0}

        Now the values have been carried over. Note that this does NOT copy the state.
        For more information on this check the documentation of the
        :meth:`init_from <qrisp.QuantumVariable.init_from>` method of QuantumVariable.
        """

        res = self.copy()

        if qs is None:
            from qrisp import QuantumSession

            qs = QuantumSession()

        indices = product(*[list(range(i)) for i in self.shape])

        for i in indices:
            temp_dup = self.qtype.duplicate(qs=qs)
            np.ndarray.__setitem__(res, i, temp_dup)

        res.msize = self.qtype.size
        res.qtype = self.qtype
        res.qs = qs
        if init:
            res.init_from(self)

        return res

    def __array_finalize__(self, obj):
        if obj is None:
            return

        self.shape_specified = bool(obj.shape_specified)
        self.qtype = obj.qtype
        self.qs = obj.qs

    def most_likely(self, **kwargs):
        """
        Performs a measurement and returns the most likely outcome.

        Parameters
        ----------
        **kwargs : Keyword arguments for the get_measurement call.

        Examples
        --------

        >>> from qrisp import QuantumFloat, QuantumArray, ry
        >>> import numpy as np
        >>> qa = QuantumArray(QuantumFloat(3), shape = 4)
        >>> ry(np.pi*9/8, qa[0][0])
        >>> print(qa)
        {OutcomeArray([1, 0, 0, 0]): 0.9619, OutcomeArray([0, 0, 0, 0]): 0.0381}
        >>> qa.most_likely()
        OutcomeArray([1, 0, 0, 0])
        """
        return list(self.get_measurement(**kwargs))[0]


@conjugate
def manipulate_array(q_array, index):
    from qrisp import QuantumFloat, demux, invert

    if isinstance(index, tuple):
        if len(q_array.shape) != len(index):
            raise Exception(
                "Tried to quantum deref QuantumArray with index of mismatching shape"
            )

        for qf in index:
            if isinstance(qf, QuantumFloat):
                if qf.signed:
                    raise Exception("Tried to quantum deref with a signed QuantumFloat")
                if qf.exponent != 0:
                    raise Exception("Tried to quantum deref with a non-integer")

        index = sum([qv.reg for qv in index[::-1]], [])
        q_array = q_array.flatten()

    with invert():
        demux(q_array[0], index, q_array)

    return q_array[0]


class OutcomeArray(np.ndarray):
    def __new__(subtype, ndarray):
        if isinstance(ndarray, list):
            ndarray = np.array(ndarray)

        obj = super().__new__(subtype, ndarray.shape, dtype=ndarray.dtype)
        indices = product(*[list(range(i)) for i in ndarray.shape])
        for i in indices:
            np.ndarray.__setitem__(obj, i, ndarray[i])

        obj.flags.writeable = False
        return obj

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other
        else:
            return np.array_equal(self, other)

    def __repr__(self):
        res = np.ndarray.__repr__(self).replace(", dtype=int64", "")

        return res



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-


import threading

import numpy as np
from numba import njit

from qrisp.circuit import Instruction, QuantumCircuit, transpile, Reset

memory_bandwidth_penalty = 2


# This class is supposed to describe a group of instructions
# The idea behind the grouping is that grouping instructions together allows
# to precalculate their unitary. This saves alot of time because applying
# a medium size unitary on a large statevector is more efficient than applying
# many small unitaries. This estimation is elaborated in the calc_gain method.
class GroupedInstruction:
    # The constructor takes a list of instruction (for instance from a quantum circuit)
    # and a list of indices, which describe which instruction to include in the group
    # Using the qubits argument, is possible to provide a list of qubits, where the
    # instruction are acting on. Otherwise, the (rather slow) .merge method is used
    def __init__(self, int_qc, indices, qubits=None):
        self.gate_signature_list = []

        if qubits == None:
            qubit_set = 0

            for i in range(len(indices)):
                qubit_set |= int_qc.data[indices[i]]

            qubit_set = int_to_qb_set(int(qubit_set), int_qc.source)
        else:
            qubit_set = set(qubits)

        self.qubits = list(qubit_set)

        self.instr_list = int_qc.source.data

        self.indices = indices

        # We now calculate the gain by estimating how many floating point operations are
        # performed for the grouped circuit vs the non-grouped circuit

        # Consider a statevector s of size 2**n and k small unitaries U_i of shape
        # (2**l, 2**l). Applying one of the unitaries to the large statevector can be
        # understood as a block matrix application
        # (U_i, 0, 0, 0)
        # (0, U_i, 0, 0)
        # (0, 0, U_i, 0)
        # (0, 0, 0, U_i)
        # i.e. 2**(n-l) applications of a (2**l, 2**l) matrix.

        # Counting the floating point operations we have 2**(n-l)*(2**l)**2 = 2**(n+l)
        # for a single matrix multiplication. Therefore, accounting for all k unitaries,
        # we have FLOPS = k*2**(n+l). If we now group the k unitaries into a medium
        # sized unitary of shape (2**L, 2**L) we therefore get a FLOP count of
        # FLOPS = 2**(n+L). Note that this assumes that calculating the medium-sized
        # unitary can be calculated for free. As it turns out, for the scales where the
        # pythonic slowdown of this algorithm does not decrease the speed, this
        # assumption is largely valid. (Check the calc_circuit_unitary function of
        # unitary_management.py)

        # Since both FLOP counts scale with 2**n, we simply calculate k*2**l
        # (or better the SUM instead of just taking the product) and 2**L

        self.gain = 0
        for i in range(len(indices)):
            self.gain += (
                1 << self.instr_list[indices[i]].op.num_qubits
            )  # + memory_bandwidth_penalty

        self.gain = self.gain - 2 ** len(self.qubits) * 0.45

    def get_instruction(self):
        # Create the QuantumCircuit which describes this group of instructions

        # Else create a quantum circuit from the given qubits
        temp_qc = QuantumCircuit()
        temp_qc.qubits = self.qubits

        for i in range(len(self.indices)):
            for cb in self.instr_list[self.indices[i]].clbits:
                try:
                    temp_qc.add_clbit(cb)
                except:
                    pass

            temp_qc.data.append(self.instr_list[self.indices[i]])
        # Create instruction
        self.instruction = Instruction(temp_qc.to_op(), temp_qc.qubits, temp_qc.clbits)

        return self.instruction


# The idea is now to iterate through different groupings and find the one with the most
# gain.
def group_qc(qc):
    # This parameter describes how deep the search for good groupings should go
    max_recursion_depth = optimal_grouping_recursion_parameter(len(qc.qubits)) + 12

    # Set up list of grouped instructions
    grouped_instr_list = []

    # We now succesively remove gates from the front of the circuit that either
    # have been determined as group or can't be grouped (because they are non-unitary)

    int_qc = IntegerCircuit(qc)
    while qc.data:
        # If the instruction is non unitary, remove
        # if qc.data[0].op.name in ["measure", "reset", "disentangle"] or
        # (qc.data[0].op.name == "pt_permutation" and qc.data[0].op.num_qubits > 7):
        if int_qc.data[0] < 0:
            grouped_instr_list.append(int_qc.source.data.pop(0))
            int_qc.data.pop(0)
            continue

        # Find the best grouping
        group = find_group(int_qc, max_recursion_depth)

        # Append the grouped instruction to the result list
        # grouped_instr_list.append(group.instruction)
        grouped_instr_list.append(group.get_instruction())

        # Delete the instructions from the circuit
        delete_multiple_element(int_qc.source.data, group.indices)
        delete_multiple_element(int_qc.data, group.indices)

    # Creat resulting circuit
    grouped_qc = qc.clearcopy()
    for i in range(len(grouped_instr_list)):
        grouped_qc.data.append(grouped_instr_list[i])

    return grouped_qc


# This function determines a set of grouping options and chooses the best option
def find_group(int_qc, max_recursion_depth):
    try:
        int_qc.data = np.array(int_qc.data, dtype=np.int64)
    except OverflowError:
        pass
    grouping_options = find_grouping_options(int_qc, [], max_recursion_depth)

    int_qc.data = list(int_qc.data)
    # print(len(grouping_options))
    return_index = np.argmax([gi.gain for gi in grouping_options])

    if grouping_options[return_index].gain > 0:
        return grouping_options[return_index]

    return GroupedInstruction(int_qc, [0])


# The groupings are determined by choosing a set of qubits and then trying which
# instructions can be executed on these qubits without "leaving" this set of qubits.
def find_grouping_options(
    int_qc, traversed_qb_sets, max_recursion_depth, qubits=None, established_indices=[]
):
    # If no set of qubit is proposed, we search for the instructions that could
    # be executed now (ie. in a QuantumCircuit they would be on the left end)
    # and search for groups starting with these instructions

    if qubits is None:
        qubits = int_qc.data[0]
        established_indices = [0]

    # Log that this qubit constellation has been traversed
    # traversed_qb_sets.append(sum([hash(qb) for qb in qubits]))
    traversed_qb_sets.append(qubits)

    # Get the indices of the instructions that can be executed without leaving the given
    # set of qubits. Additionally, this function returns the list expansion_options:
    # This list contains qubits that could be included in the group
    # instruction_indices, expansion_options = get_circuit_block(int_qc.source.data,
    # int_to_qb_set(qubits, int_qc.source), list(established_indices))

    instruction_indices, expansion_options = get_circuit_block_(
        int_qc, qubits, np.array(established_indices)
    )

    # Create the list of grouping options
    # options = [GroupedInstruction(qc.source.data, instruction_indices,  qubits)]
    options = [
        GroupedInstruction(
            int_qc, instruction_indices, int_to_qb_set(qubits, int_qc.source)
        )
    ]

    # Check some conditions that determine if no further recursion should be performed
    if (
        len(expansion_options) == 0
        or max_recursion_depth == 0
        or len(int_to_qb_set(qubits, int_qc.source)) >= 7
    ):
        return options

    # Now, include the proposed expansion options
    for i in range(len(expansion_options)):
        # Calculate the hash of the proposed set of qubits
        # proposed_set = sum([hash(qb) for qb in qubits + [expansion_options[i]]])
        proposed_set = qubits | qb_set_to_int([expansion_options[i]], int_qc.source)
        # proposed_set = qubits.union(BinaryQubitSet([expansion_options[i]], qc.source))

        # If this set has not been checked yet, add to the options
        if not proposed_set in traversed_qb_sets:
            # options += find_grouping_options(qc_data, traversed_qb_sets,
            # max_recursion_depth - 1, qubits = qubits + [expansion_options[i]],
            # established_indices = list(instruction_indices))
            options += find_grouping_options(
                int_qc,
                traversed_qb_sets,
                max_recursion_depth - 1,
                qubits=proposed_set,
                established_indices=(instruction_indices),
            )

    return options


# This function returns the indices of the instructions of a circuit,
# where the returned can be executed by only "staying inside" a given set of qubits
def get_circuit_block(qc_data, qubits, established_indices=[]):
    # Copy set in order to prevent modification
    qubits = set(qubits)

    # Set up set of expansion options
    expansion_options = set([])

    # Set up result list
    instruction_indices = []

    # Now iterate through all the given instructions
    for i in range(len(qc_data)):
        if len(qubits) == 0:
            break

        # If the instruction has been identified as part of the group
        # in a previous recursion, skip the checking and add to the list
        # of instructions
        if len(established_indices):
            if established_indices[0] == i:
                instruction_indices.append(established_indices.pop(0))
                continue

        # Set some aliases
        instr = qc_data[i]
        instr_qubits = set(instr.qubits)

        # Determine the intersection between the qubits of the instruction
        # and the qubits of the group
        intersection = instr_qubits.intersection(qubits)

        # If the intersection is empty, this instruction is not part of the group
        # and no further action needs to be taken
        if not intersection:
            continue

        # If the instruction is non-unitary, no further instruction
        # on this qubit can be part of the group
        if instr.op.name in ["measure", "reset", "disentangle"]:
            qubits = qubits - instr_qubits
            continue

        # If the instruction qubits are part of the group qubits,
        # add the instruction to the group
        if instr_qubits.issubset(qubits):
            instruction_indices.append(i)

        # Otherwise, the instruction happens partly on the group qubits,
        # partly outside. Therefore, we need to remove the qubits
        # that interact with the outside.
        # Nevertheless, we add the "outside" qubits to the set of expansion options
        else:
            qubits = qubits - intersection
            expansion_options = expansion_options.union(instr_qubits - intersection)

    # Return result
    return instruction_indices, list(expansion_options)


# Empirically determined parameters that seem to work best
def optimal_grouping_recursion_parameter(qubit_amount):
    if qubit_amount <= 16:
        return 2
    elif 16 < qubit_amount <= 20:
        return 3
    elif 20 < qubit_amount <= 24:
        return 4
    elif 24 < qubit_amount <= 28:
        return 6
    elif 28 < qubit_amount <= 32:
        return 7
    elif 32 < qubit_amount < 35:
        return 8
    else:
        return 9


# This function inserts disentangling operations into the circuit, if suited
# Consider the following circuit
#             ┌───┐
# qubit_3362: ┤ H ├──■──────────────────────
#             └───┘┌─┴─┐┌───┐
# qubit_3363: ─────┤ x ├┤ H ├──■────────────
#                  └───┘└───┘┌─┴─┐┌────────┐
# qubit_3364: ───────────────┤ x ├┤ P(0.3) ├
#                            └───┘└────────┘
# If we don't the gates after the first CNOT with U,
# the resulting state is
# 2**-0.5*(|0>U|0> + |1>U|1>)
# Since we are not interested in the wave-function but in the resulting
# probabilities, it is now safe to evaluate both summands of the wave function
# separately, ie. evaluate |0>U|0> and |1>U|1> and multiply the resulting,
# probabilities with 2**-0.5.
# This is possible because there is not going to be any interaction between the to
# states.
# (TO-DO better proof required)
# Therefore the workflow for performing the disentangling is
# 1. Identify position to insert disentangler (here: directly behind the first CNOT)
# 2. Perform measurement (to acquire classical probabilities)
# 3. Continue the simulation on the two decoherent states

# Simulating both states separately has two advantages
# 1. If one of the branches has probability zero, this will show up in the measurement
#   and we don't have to continue the simulation of this branch
# 2. Simulating two states that don't interact is easier to parallelize

# As it turns out, it is even possible to disentangle this circuit right
# after the first H gate on qubit zero


#              ┌───┐     ┌────────┐
# qubit_10119: ┤ H ├──■──┤ P(0.2) ├─────────────────■──
#              └───┘┌─┴─┐└─┬───┬──┘                 │
# qubit_10120: ─────┤ x ├──┤ H ├─────■──────────────┼──
#                   └───┘  └───┘   ┌─┴─┐┌────────┐┌─┴─┐
# qubit_10121: ────────────────────┤ x ├┤ P(0.3) ├┤ x ├
#                                  └───┘└────────┘└───┘


# This is because all the gates after the H-gate are permeable on this qubit
# (for an elaboration on permeability check the uncomputation module)
# Roughly said, a gate U which is permeable on the first qubit behaves like this
# U|0>|a> = exp(i*phi_0) |0> U_0 |a>
# U|1>|b> = exp(i*phi_1) |0> U_1 |b>


def insert_disentangling(qc):
    # This function checks for permeability on a given qubit
    from qrisp import Instruction, Operation
    from qrisp.uncomputation import is_permeable

    # Create disentangling operation
    disentangler = Operation("disentangle", num_qubits=1)
    disentangler.permeability = {0: False}

    # After all the operations have been performed on a qubit,
    # it can be reset without changing the statistics
    for i in range(len(qc.qubits)):
        qc.reset(qc.qubits[i])
        # qc.append(disentangler, [qc.qubits[i]])

    # return qc

    # We now insert the disentanglers at every position which
    # is only separated by permeable operations to the final operation
    # of that qubit
    reversed_data = list(qc.data)[::-1]

    disentangling_counter = 0
    i = 0
    while i < len(reversed_data):
        if not reversed_data[i].op.name in ["measure", "reset"]:
            i += 1
            continue

        qubit = reversed_data[i].qubits[0]
        j = int(i)

        while j < len(reversed_data):
            instr = reversed_data[j]
            if qubit in instr.qubits:
                if instr.op.name in ["measure", "reset", "disentangle"]:
                    j += 1
                    continue

                # Find qubit index
                qubit_index = instr.qubits.index(qubit)

                if is_permeable(instr.op, [qubit_index]):
                    # If so, insert disentangler before this instruction
                    reversed_data.insert(j + 1, Instruction(disentangler, [qubit]))
                    disentangling_counter += 1
                    j += 1

                else:
                    break
            j += 1

        i += 1

    # #For every qubit, we traverse the circuit from the back,
    # #until we find an operation which is not permeable on that qubit
    # for i in range(len(qc.qubits)):

    #     #Set qubit alias
    #     qb = qc.qubits[i]
    #     j = -1
    #     while j < len(reversed_data)-1:
    #         j += 1

    #         #Set alias for instruction that is to be checked
    #         instr = reversed_data[j]

    #         #If the instruction is a reset, measurement or disentangling,
    #         #it performs the same operation as a disentangler, therefore
    #         #we can safely insert further disentanlger before it
    #         if instr.op.name in ["measure", "reset", "disentangle"]:
    #             continue

    #         #If the instruction acts on the qubit in question,
    #         #check if the instruction is permeable on this qubit
    #         if qb in instr.qubits:

    #             #Find qubit index
    #             qubit_index = instr.qubits.index(qb)

    #             #Check for permeability
    #             if is_permeable(instr.op, [qubit_index]):

    #                 #If so, insert disentangler before this instruction
    #                 reversed_data.insert(j+1, Instruction(disentangler, [qb]))
    #                 disentangling_counter += 1
    #             #Otherwise, we can no longer insert disentanglers
    #             else:
    #                 break

    # Create output quantum circuit
    new_qc = qc.clearcopy()
    new_qc.data = reversed_data[::-1]

    return new_qc


# Simple function to count the amount of measurements in a circuit,
# this is helpfull, because we don't need to continue the simulation,
# once all measurements have been performed
def count_measurements_and_treat_alloc(qc, insert_reset=True):
    counter = 0
    i = 0

    while i < len(qc.data):
        instr = qc.data[i]

        if instr.op.name == "barrier":
            qc.data.pop(i)
            continue

        if instr.op.name == "measure":
            counter += 1
        elif instr.op.name == "qb_alloc":
            pass
            qc.data.pop(i)
            continue
        elif instr.op.name == "qb_dealloc":
            qc.data.pop(i)
            if insert_reset:
                qc.data.insert(i, Instruction(Reset(), qubits = instr.qubits))

        i += 1
    return counter


def delete_multiple_element(list_object, indices):
    indices = sorted(indices, reverse=True)
    for idx in indices:
        if idx < len(list_object):
            list_object.pop(idx)


def qb_set_to_int(qubits, qc):
    res = 0
    for qb in qubits:
        res |= 1 << qc.qubits.index(qb)
    return res


def qc_to_int_list(qc):
    res_list = []
    for instr in qc.data:
        res_list.append(qb_set_to_int(instr.qubits, qc))
        if instr.op.name in ["measure", "reset", "disentangle"] or (
            instr.op.name.find("c_if") != -1
        ):
            res_list[-1] *= -1

    return res_list


# Copy set in order to prevent modification
# qubits = set(qubits)
def get_circuit_block_(int_qc, qubits, established_indices=[]):
    # qubits = qb_set_to_int(qubits, qc.source)
    n = int_qc.n

    # If the qubits int is to large to fit into np.int64, we use the non jitted version
    if isinstance(int_qc.data, np.ndarray):
        instruction_indices, expansion_options = binary_get_circuit_block_jitted(
            int_qc.data, qubits, n, established_indices
        )
    else:
        instruction_indices, expansion_options = binary_get_circuit_block(
            int_qc.data, qubits, n, np.array(established_indices, dtype=np.int64)
        )

    return instruction_indices, int_to_qb_set(expansion_options, int_qc.source)
    # return instruction_indices, expansion_options.to_qubit_list()


@njit
def binary_get_circuit_block_jitted(int_qc_list, qubits, n, established_indices):
    # Set up set of expansion options
    expansion_options = 0

    # Set up result list
    instruction_indices = []
    ee_counter = 0
    # Now iterate through all the given instructions
    for i in range(len(int_qc_list)):
        if qubits == 0:
            break

        # If the instruction has been identified as part of the group
        # in a previous recursion, skip the checking and add to the list
        # of instructions
        if ee_counter < len(established_indices):
            if established_indices[ee_counter] == i:
                ee_counter += 1
                instruction_indices.append(i)
                continue

        # Set some aliases
        instr_qubits = int_qc_list[i]
        # instr = qc.data[i]
        # instr_qubits = set(instr.qubits)

        # Determine the intersection between the qubits of the instruction
        # and the qubits of the group
        # intersection = instr_qubits.intersection(qubits)
        intersection = qubits & instr_qubits

        # If the intersection is empty, this instruction is not part of the group
        # and no further action needs to be taken
        if not intersection:
            continue

        # If the instruction is non-unitary, no further instruction
        # on this qubit can be part of the group
        if instr_qubits < 0:
            qubits = qubits & (((1 << n) - 1) ^ instr_qubits)
            # qubits = qubits - instr_qubits
            continue

        # If the instruction qubits are part of the group qubits,
        # add the instruction to the group
        if not bool((((1 << n) - 1) ^ qubits) & instr_qubits):
            # if instr_qubits.issubset(qubits):
            instruction_indices.append(i)

        # Otherwise, the instruction happens partly on the group qubits,
        # partly outside. Therefore we need to remove the qubits
        # that interact with the outside.
        # Nevertheless, we add the "outside" qubits to the set of expansion options
        else:
            qubits = qubits & (((1 << n) - 1) ^ intersection)
            # qubits = qubits - intersection
            expansion_options = expansion_options | (
                instr_qubits & (((1 << n) - 1) ^ intersection)
            )
            # expansion_options = expansion_options.union(instr_qubits - intersection)

    # Return result
    return instruction_indices, expansion_options


def binary_get_circuit_block(int_qc_list, qubits, n, established_indices):
    # Set up set of expansion options
    expansion_options = 0

    # Set up result list
    instruction_indices = []
    ee_counter = 0
    # Now iterate through all the given instructions
    for i in range(len(int_qc_list)):
        if qubits == 0:
            break

        # If the instruction has been identified as part of the group
        # in a previous recursion, skip the checking and add to the list
        # of instructions
        if ee_counter < len(established_indices):
            if established_indices[ee_counter] == i:
                ee_counter += 1
                instruction_indices.append(i)
                continue

        # Set some aliases
        instr_qubits = int_qc_list[i]
        # instr = qc.data[i]
        # instr_qubits = set(instr.qubits)

        # Determine the intersection between the qubits of the instruction
        # and the qubits of the group
        # intersection = instr_qubits.intersection(qubits)
        intersection = qubits & instr_qubits

        # If the intersection is empty, this instruction is not part of the group
        # and no further action needs to be taken
        if not intersection:
            continue

        # If the instruction is non-unitary, no further instruction
        # on this qubit can be part of the group
        if instr_qubits < 0:
            qubits = qubits & (((1 << n) - 1) ^ instr_qubits)
            # qubits = qubits - instr_qubits
            continue

        # If the instruction qubits are part of the group qubits,
        # add the instruction to the group
        if not bool((((1 << n) - 1) ^ qubits) & instr_qubits):
            # if instr_qubits.issubset(qubits):
            instruction_indices.append(i)

        # Otherwise, the instruction happens partly on the group qubits,
        # partly outside. Therefore we need to remove the qubits
        # that interact with the outside.
        # Nevertheless, we add the "outside" qubits to the set of expansion options
        else:
            qubits = qubits & (((1 << n) - 1) ^ intersection)
            # qubits = qubits - intersection
            expansion_options = expansion_options | (
                instr_qubits & (((1 << n) - 1) ^ intersection)
            )
            # expansion_options = expansion_options.union(instr_qubits - intersection)

    # Return result
    return instruction_indices, expansion_options


def int_to_qb_set(integer, qc):
    res = []
    for i in range(len(qc.qubits)):
        if int(integer) & (1 << i):
            res.append(qc.qubits[i])

    return res


class IntegerCircuit:
    def __init__(self, qc):
        self.source = qc
        self.data = qc_to_int_list(qc)
        self.n = len(qc.qubits)


def average_group_size(qc):
    average_group_size = 0

    for i in range(len(qc.data)):
        if qc.data[i].op.definition:
            average_group_size += len(qc.data[i].op.definition.data)
        else:
            average_group_size += 1

    return average_group_size / len(qc.data)


# Wrapping function for all preproccessing operations
def circuit_preprocessor(qc):
    # from qrisp.simulator.permutation_identification import group_pt_permutations
    from qrisp.simulator import reorder_circuit

    if len(qc.data) == 0:
        return qc.copy()

    # TO-DO find reliable classifiaction when automatic disentangling works best
    if len(qc.qubits) < 26:
        qc = group_qc(qc)
    elif len(qc.qubits) < 34:
        qc = group_qc(qc)
        qc = insert_disentangling(qc)
    else:
        qc = insert_disentangling(qc)
        qc = group_qc(qc)

    return reorder_circuit(qc, ["measure", "reset", "disentangle"])



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

# The functions in this file are supposed to reorder circuits in such a way,
# that measurements, resets and disentanglers are performed as early as possible.
# This is beneficial because the simulation of two decoherent states can be parallelized
# better. Furthermore, might the measurement of these imply that one of the states has
# vanishing probability, which means that this state does not require further
# simulation.

# The idea to achieve the ordering is to convert the circuit into a directed acyclic
# graph (now called causal graph), where two successive operations with overlapping
# qubits are represented with two nodes, connected by a directed edge. The directed
# edge, however points into the opposite direction of the sequential order of the
# operations. This way, if we want to evaluate what gates are necessary in order to
# perform a measurement, disentangling or reset, we simply have to look at the set of
# gates that is reachable from that specific node in the causal graph.

# Consider the following circuit
#           ┌───┐
#  qubit_9: ┤ X ├──■─────
#           ├───┤┌─┴─┐┌─┐
# qubit_10: ┤ Y ├┤ x ├┤M├
#           ├───┤└───┘└╥┘
# qubit_11: ┤ H ├──────╫─
#           └───┘      ║
#  clbit_0: ═══════════╩═

# In order to perform the measurement, we have to execute the x and the cx gate
# The corresponding causal graph
# (measure)-> (CX)->(X)
#               L->(Y)
# (H)
# We see that CX, X and Y are reachable from the measurement
# We now perform a depth first traversal on the subgraph of the nodes which are
# reachable from the measurement.

# It is important to node that for a regular depth first traversal,
# we are free to choose which child of a node to traverse next.
# This is not the casehere, because this might disturb the topological
# order, which would imply that we performed an illegitimat reordering
# Therefore, in this depth first traversal, we always traverse the child
# with the highest index in a topological order first.

import networkx as nx
import numpy as np


def nx_get_causal_graph(
    qc, inverted=False, get_non_unitary_nodes=False, preferential_gates=[]
):
    from qrisp.circuit import Instruction, Operation

    # Create graph object
    G = nx.DiGraph()

    # This distionary contains the information, which node of the graph
    # if the most up to date noce for a qubit
    current_node_qubits = {}

    # List to collect the non-unitary nodes
    non_unitary_nodes = []

    # Now we traverse the circuit instructions
    for i in range(len(qc.data)):
        # Add new node to the graph
        new_node = i
        G.add_node(new_node)

        # Determine the set of nodes, which the new node will connect to
        # ie. the nodes which are most up to date on the particular qubits
        node_set = []
        for qb in qc.data[i].qubits:
            # If there is a node which has been connected to this qubit before,
            # append the node to the node set, that the new node will be connected to
            try:
                node_set.append(current_node_qubits[qb])
            except KeyError:
                pass

            # Update the dictionary
            current_node_qubits[qb] = new_node

        # Make sure every node is listed only once
        node_set = list(set(node_set))

        # Add the edges
        if inverted:
            for node in node_set:
                G.add_edge(new_node, node)
        else:
            for node in node_set:
                G.add_edge(node, new_node)

        # Log if the new node is non unitary
        if qc.data[i].op.name in preferential_gates + ["final_op"]:
            non_unitary_nodes.append(new_node)

    # Return result
    if get_non_unitary_nodes:
        return G, non_unitary_nodes

    return G


# Function to reorder a circuit as described above
def nx_reorder_circuit(qc, preferential_gates=[]):
    from qrisp.circuit import Instruction, Operation

    for i in range(len(qc.qubits)):
        qc.append(Operation("final_op", num_qubits=1), [qc.qubits[i]])

    # Acquire causal graph
    G, non_unitary_nodes = nx_get_causal_graph(
        qc,
        inverted=True,
        get_non_unitary_nodes=True,
        preferential_gates=preferential_gates,
    )

    from networkx import descendants, topological_sort, transitive_reduction

    # We now order the non-unitary nodes according to how many decendents they all have
    # Measurements/Resets/Disentanglings with only a few descendants need only a few
    # gates to be simulated until the measurement can be executed
    node_costs = []
    for i in range(len(non_unitary_nodes)):
        if qc.data[hash(non_unitary_nodes[i])].op.name == "final_op":
            node_costs.append((non_unitary_nodes[i], np.inf))
        else:
            node_costs.append(
                (non_unitary_nodes[i], len(descendants(G, non_unitary_nodes[i])))
            )

    node_costs.sort(key=lambda x: x[1])

    # We now determine the topological dictionary of the nodes.
    # This dictionary assigns each node an integer, which indicates its position in a
    # topological ordering. The topological ordering of a graph has the feature that two
    # nodes N1, N2 which stand in a causal relationship to each other (i.e. N1 has to be
    # executed after N2) also have increasing index in the topological ordering i.e.
    # tp_dic[N1] > tp_dic[N2]
    sorted_nodes = list(topological_sort(G))
    tp_dic = {sorted_nodes[i]: i for i in range(len(sorted_nodes))}

    # This function now performs depth first traversal of the given causal graph,
    # starting at node. Each time a node is visited the callback function is called

    # The callback function that this will mainly be used with, is a logger,
    # which notes in which order the nodes have been visited

    # As mentioned in the comment at the beginning of this file,
    # a regular depth first traversal basically allows picking any child
    # to traverse next. In our case, this can mess with the topological ordering
    # Therefore we allways traverse the child with the highest topological index

    def topological_desc_traversal(G, node, tp_dic, callback):
        node_list = [x for x in nx.descendants(G, node)]

        node_list.sort(key=lambda x: -tp_dic[x])

        for n in node_list:
            callback(n)
            G.remove_node(n)

        callback(node)
        G.remove_node(node)
        return

    def topological_df_traversal(G, node, tp_dic, callback):
        # Acquire list of children
        node_list = [x for x in G.neighbors(node)]

        # Sort according to their topological index
        node_list.sort(key=lambda x: -tp_dic[x])

        # Recursively traverse the children
        while node_list:
            topological_df_traversal(G, node_list[0], tp_dic, callback)

            # Since the traversed nodes are removed after each traversal,
            # we need to to update the list of children, in case any of
            # the nodes in node_list have been removed
            node_list = [x for x in G.neighbors(node)]
            node_list.sort(key=lambda x: -tp_dic[x])

        # Call the callback function
        callback(node)
        # Remove the node in order to prevent traversing it again
        # this can happen because the causal graph is not necessarily a tree graph
        G.remove_node(node)

    # The circuits in this list will be the circuits whose execution is the absolute
    # minimum in order to evaluate a certain non-unitary operation
    new_qc_list = []

    # Now we succesively determine the minimal circuit required to execute
    # for each non-unitary operation
    while node_costs:
        # This node contains the non-unitary operation
        evaluation_node = node_costs.pop(0)[0]

        # This list will contain the unitary operations that are necessary
        # in order to perform the operation described by evaluation_node
        evaluation_list = []

        # Create callback function
        def callback(x):
            evaluation_list.append(x)

        # Traverse causal graph
        topological_desc_traversal(G, evaluation_node, tp_dic, callback)
        # topological_df_traversal(G, evaluation_node, tp_dic, callback)

        # Create circuit
        new_qc = qc.clearcopy()
        # Append the corresponding instruction to the circuit
        for node in evaluation_list:
            instr = qc.data[hash(node)]
            if not instr.op.name == "final_op":
                new_qc.data.append(instr)

        new_qc_list.append(new_qc)

    # Create result quantum circuit
    new_qc = qc.clearcopy()

    # Concatenate the data of the newly created circuits
    new_qc.data = sum([qc.data for qc in new_qc_list], [])

    # Remove final_op operations
    for i in range(len(qc.qubits)):
        qc.data.pop(-1)

    # Return result
    return new_qc


# Similar function as above but implemented for the C++ based
# graph theory package networkit
def nk_reorder_circuit(qc, preferential_gates=[]):
    from qrisp.circuit import Instruction, Operation

    for i in range(len(qc.qubits)):
        qc.append(Operation("final_op", num_qubits=1), [qc.qubits[i]])

    import networkit as nk

    G = nk.Graph(directed=True)
    current_node_qubits = {qubit: "-" for qubit in qc.qubits}
    non_unitary_nodes = []
    measurement_counter = 0
    for i in range(len(qc.data)):
        new_node = G.addNode()
        node_set = []
        for qb in qc.data[i].qubits:
            if current_node_qubits[qb] != "-":
                node_set.append(current_node_qubits[qb])

            current_node_qubits[qb] = new_node
        node_set = list(set(node_set))

        for node in node_set:
            G.addEdge(i, node)

        if qc.data[i].op.name in preferential_gates + ["final_op"]:
            non_unitary_nodes.append(new_node)
            if qc.data[i].op.name == "measure":
                measurement_counter += 1

    from networkit.traversal import Traversal

    new_qc_list = []

    from networkit.graphtools import GraphTools

    sorted_nodes = GraphTools.topologicalSort(G)

    tp_dic = {sorted_nodes[i]: i for i in range(len(sorted_nodes))}

    reach_alg = nk.reachability.ReachableNodes(G, exact=True)

    reach_alg.run()
    node_costs = []

    for i in range(len(non_unitary_nodes)):
        if qc.data[hash(non_unitary_nodes[i])].op.name == "final_op":
            node_costs.append((non_unitary_nodes[i], np.inf))
        else:
            node_costs.append(
                (
                    non_unitary_nodes[i],
                    reach_alg.numberOfReachableNodes(non_unitary_nodes[i]),
                )
            )

    node_costs.sort(key=lambda x: x[1])

    def topological_desc_traversal(G, node, tp_dic, callback):
        node_list = []

        nk.traversal.Traversal.DFSfrom(G, node, node_list.append)

        node_list.sort(key=lambda x: -tp_dic[x])

        for n in node_list:
            callback(n)
            G.removeNode(node)

        # callback(node)
        # G.removeNode(node)

        return

    def topological_df_traversal(G, node, tp_dic, callback):
        node_list = [x for x in G.iterNeighbors(node)]
        node_list.sort(key=lambda x: -tp_dic[x])

        while node_list:
            topological_df_traversal(G, node_list[0], tp_dic, callback)
            node_list = [x for x in G.iterNeighbors(node)]
            node_list.sort(key=lambda x: -tp_dic[x])

        callback(node)
        G.removeNode(node)

    while node_costs:
        evaluation_node = node_costs.pop(0)[0]

        evaluation_list = []

        def callback(x, y=0):
            if len(evaluation_list):
                G.removeNode(evaluation_list[-1])
            evaluation_list.append(x)

        new_qc = qc.clearcopy()

        # topological_df_traversal(G, evaluation_node, tp_dic, callback)
        topological_desc_traversal(G, evaluation_node, tp_dic, callback)

        for node in evaluation_list:
            instr = qc.data[hash(node)]
            if instr.op.name != "final_op":
                new_qc.data.append(instr)

        new_qc_list.append(new_qc)

    for i in range(len(qc.qubits)):
        qc.data.pop(-1)

    new_qc = qc.clearcopy()
    new_qc.data = sum([qc.data for qc in new_qc_list], [])

    return new_qc


try:
    import networkit

    nk_available = True

except:
    nk_available = False
    # print("Install networkit for additional simulator performance.
    # Resorting to networkx.")


def reorder_circuit(qc, preferential_gates=[]):
    if nk_available:
        # return nk_reorder_circuit(qc, preferential_gates)

        return nx_reorder_circuit(qc, preferential_gates)
    else:
        return nx_reorder_circuit(qc, preferential_gates)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 06.05.2022
import numpy as np

from qrisp.misc import multi_measurement
from qrisp.core import QuantumSession
from qrisp.arithmetic import QuantumFloat, sbp_mult
from qrisp.grover import tag_state, grovers_alg
from qrisp.environments import invert
from itertools import product
from qrisp import p, QuantumVariable, h, mcx, QuantumBool


def test_grovers_algorithm():
    qf1 = QuantumFloat(2)
    qf2 = QuantumFloat(2)

    # Put into list
    qf_list = [qf1, qf2]

    # Create test oracle (simply tags the states where qf1 = -3 and qf2 = 2)
    def test_oracle(qf_list):
        tag_dic = {qf_list[0]: 0, qf_list[1]: 1}
        tag_state(tag_dic)

    # Apply grovers algorithm
    grovers_alg(qf_list, test_oracle)

    print(qf_list[0].qs)
    # Perform measurement
    mes_res = multi_measurement(qf_list)
    print(mes_res)

    # check if it is dictionary
    assert isinstance(mes_res, dict)

    # check if the key value pairs are of the type tuple:float
    # Example: (0, 1) :  0.9592
    assert all(isinstance(item, tuple) for item in mes_res.keys()) and all(
        (isinstance(item, float) and item < 1) for item in mes_res.values()
    )

    # Check measurement outcome
    assert list(mes_res.keys())[0] == (0, 1)

    # check if the first tuple in the tuple contains expected values. Style example of the whole tuple: ((0, 1), 0.9592)
    assert all(
        (item in [ele for ele in product(range(0, 4), repeat=2)])
        for item in mes_res.keys()
    )

    ###################
    # Create quantum floats
    qf1 = QuantumFloat(2, -1, signed=True)
    qf2 = QuantumFloat(2, -1, signed=True)

    # Put into list
    qf_list = [qf1, qf2]

    # Create test oracle (tags the states where the multiplication results in -0.25)
    def equation_oracle(qf_list):
        # Set aliases
        factor_0 = qf_list[0]
        factor_1 = qf_list[1]

        # Calculate ancilla value
        ancilla = factor_0 * factor_1

        # Tag state
        tag_dic = {ancilla: -0.25}
        tag_state(tag_dic)

        # Uncompute ancilla
        with invert():
            sbp_mult(factor_0, factor_1, ancilla)

        ancilla.delete()

    # Execute grovers algorithm
    grovers_alg(qf_list, equation_oracle, winner_state_amount=2)

    # Perform measurement

    mes_res = multi_measurement(qf_list)
    print(mes_res)

    # check if it is a list of dictionary
    assert isinstance(mes_res, dict)

    # check if the key value pairs are of the type tuple:float
    # Example: (0, 1) :  0.9592
    assert all(isinstance(item, tuple) for item in mes_res.keys()) and all(
        (isinstance(item, float) and item < 1) for item in mes_res.values()
    )

    assert list(mes_res.keys())[0] == (0.5, -0.5)
    assert list(mes_res.keys())[1] == (-0.5, 0.5)

    # check if the first tuple in the tuple contains expected values. Style example of the whole tuple: ((0, 1), 0.9592)
    assert all(
        (
            item
            in [
                ele
                for ele in product(
                    [0.0, -0.5, 0.5, -1.0, 1.0, -1.5, 1.5, -2.0, 2.0], repeat=2
                )
            ]
        )
        for item in mes_res.keys()
    )

    # Test exact Grovers alg

    def oracle(qv, phase=np.pi):
        temp_qbl = QuantumBool()
        mcx(qv[1:], temp_qbl)

        p(phase, temp_qbl)

        temp_qbl.uncompute()

    qv = QuantumVariable(6)
    grovers_alg(qv, oracle, exact=True, winner_state_amount=2)

    assert qv.get_measurement() == {"011111": 0.5, "111111": 0.5}



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 29.04.2022
# import pytest
import numpy

from qrisp.interface import VirtualQiskitBackend
from qrisp import (
    QuantumFloat,
    QuantumChar,
    QuantumArray,
    merge,
    x,
    h,
    OutcomeArray,
    multi_measurement,
    auto_uncompute,
    invert
)


# check if the respone from QuantumArray.get_measurement() has the expected formet.
# Should return a list of tuples of the type (numpy.ndarray, float)
# ie. [(array([1,1,0]), 232), (array([1,1,3]), 115), ...]
def test_conditional_environments_example():
    qc = QuantumChar(name="q ch")

    qc[:] = "f"

    qf = QuantumFloat(4, -1, signed=True)
    q_array = QuantumArray(qf, 4)

    h(qc[0])

    q_array[0] += -0.5

    with "f" == qc:
        h(q_array[0][0])

        with q_array[0] == -0.5:
            q_array[2].encode(-0.5)
            q_array[1] -= -1
            with q_array[1] == 1:
                q_array[3].init_from(q_array[2])

        # Test in-environment printing
        print(qc.qs)

    print(multi_measurement([qc, q_array]))
    assert multi_measurement([qc, q_array]) == {
        ("e", OutcomeArray([-0.5, 0.0, 0.0, 0.0])): 0.5,
        ("f", OutcomeArray([-1.0, 0.0, 0.0, 0.0])): 0.25,
        ("f", OutcomeArray([-0.5, 1.0, -0.5, -0.5])): 0.25,
    }

    assert len(qc.qs.qv_list) == 5

    from qrisp import as_hamiltonian, QuantumBool, control

    @as_hamiltonian
    def dummy_d_function(i, j):
        return i * j

    def dummy_s_function(qf_list):
        m = len(qf_list)
        for i in range(m):
            dummy_d_function(qf_list[i], qf_list[(i + 1) % m])

    qf_list = [QuantumFloat(2) for j in range(3)]

    qf_list[-1][:] = 1

    qbool = QuantumBool()
    h(qbool)
    with control(qbool[0]):
        dummy_s_function(qf_list)
        print(qbool.qs)

    h(qbool)

    assert qbool.most_likely() == False

    # -------

    a = QuantumFloat(2, name="a")
    b = QuantumFloat(2, name="b")
    c = QuantumFloat(1, name="c")
    d = QuantumFloat(2, name="d")

    a[:] = 1
    b[:] = 1
    c[:] = 1

    def test_f(a):
        return a == 1

    from qrisp import ConditionEnvironment, ControlEnvironment

    with test_f(a):
        with b == c:
            x(d)
            print(d.qs)

    # Test if uncomputation worked properly
    assert len(a.qs.qv_list) == 4

    # Test if condition environments get invoked properly
    with b == 2:
        assert isinstance(b.qs.env_stack[-1], ConditionEnvironment)
        temp = d == 1
        # print(d.qs)
        with temp:
            assert isinstance(d.qs.env_stack[-1], ControlEnvironment)
            x(c)
            # print(c.qs)

    assert len(a.qs.qv_list) == 5

    # ----------
    a = QuantumFloat(2, signed=True)
    b = QuantumFloat(2, signed=True)
    c = QuantumFloat(1, signed=True)
    d = QuantumFloat(2, signed=True)

    a[:] = 1
    b[:] = 1
    c[:] = 1

    def test_f(a):
        return a == 1

    with test_f(a):
        with b == c:
            x(d)
            print(d.qs)

    assert len(a.qs.qv_list) == 4
    
    @auto_uncompute
    def test_function(qf_a, qbl):
        
        with qf_a == 0:
            x(qbl)
        
        # print(qbl.qs)
        return qbl
        
    qf_a = QuantumFloat(2)
    qbl = QuantumBool()

    # qf_a[:] = 3

    with invert():
        test_function(qf_a, qbl)
        
    print(qf_a.qs)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


# This file uses the gate constructor from operation.py to define
# some standard operations

import types

import numpy as np

from qrisp.circuit import PauliGate, U3Gate
from qrisp.circuit.operation import Operation

# TODO: properly treat all gates
# from Aer.get_backend('qasm_simulator').configuration().basis_gates


def Measurement():
    res = Operation(name="measure", num_qubits=1, num_clbits=1)
    res.permeability = {0: True}
    return res


def XGate():
    return PauliGate(name="x")


def YGate():
    return PauliGate(name="y")


def ZGate():
    return PauliGate(name="z")


def CXGate():
    return XGate().control()


def CYGate():
    return YGate().control()


def CZGate():
    return ZGate().control()


def MCXGate(control_amount=1, ctrl_state=-1, method="gray"):
    return XGate().control(control_amount, method=method, ctrl_state=ctrl_state)


def PGate(phi=0):
    res = U3Gate(0, 0, phi, name="p")
    return res


def CPGate(phi=0):
    if isinstance(phi, (int, float)):
        phi = phi % (2 * np.pi)
        if np.abs(phi - np.pi) < 1e-8:
            return CZGate()
        elif np.abs(phi) < 1e-8:
            from qrisp.circuit.quantum_circuit import QuantumCircuit

            temp_circ = QuantumCircuit(2)
            res = temp_circ.to_gate(name="cp")
            res.params = [0]
            return res

        else:
            return PGate(phi).control()
    else:
        return PGate(phi).control()


def HGate():
    res = U3Gate(np.pi / 2, 0, np.pi, name="h")
    return res


def RXGate(phi=0):
    res = U3Gate(phi, -np.pi / 2, np.pi / 2, name="rx")
    return res


def RYGate(phi=0):
    res = U3Gate(phi, 0, 0, name="ry")
    return res


def RZGate(phi=0):
    res = U3Gate(0, phi, 0, name="rz", global_phase=-phi / 2)
    return res


def GPhaseGate(phi=0):
    res = U3Gate(0, 0, 0, name="gphase", global_phase=phi)
    res.params = [phi]
    res.permeability[0] = True
    res.is_qfree = True
    return res


def MCRXGate(phi=0, control_amount=0):
    res = RXGate(phi).control(control_amount)
    res.name = "mcrx"
    return res


def SGate():
    res = PGate(np.pi / 2)
    res.params = []
    res.name = "s"
    return res


def TGate():
    res = PGate(np.pi / 4)
    res.params = []
    res.name = "t"
    return res


def IDGate():
    res = U3Gate(0, 0, 0, name="id")
    return res


def RXXGate(phi=0):
    from qrisp.circuit.quantum_circuit import QuantumCircuit

    qc = QuantumCircuit(2)
    qc.gphase(-phi / 2, 0)
    qc.h(0)
    qc.h(1)
    qc.cx(0, 1)
    qc.p(phi, 1)
    qc.cx(0, 1)
    qc.h(0)
    qc.h(1)

    return Operation(
        name="rxx", num_qubits=2, num_clbits=0, params=[phi], definition=qc
    )


def RZZGate(phi=0):
    from qrisp.circuit.quantum_circuit import QuantumCircuit

    qc = QuantumCircuit(2)
    qc.gphase(-phi / 2, 0)
    qc.cx(0, 1)
    qc.p(phi, 1)
    qc.cx(0, 1)

    return Operation(
        name="rzz", num_qubits=2, num_clbits=0, params=[phi], definition=qc
    )

def XXYYGate(phi=0, beta=0):
    from qrisp.circuit.quantum_circuit import QuantumCircuit

    qc = QuantumCircuit(2)
    qc.rz(beta, 0)
    qc.rz(-np.pi/2, 1)
    qc.sx(1)
    qc.rz(np.pi/2, 1)
    qc.s(0)
    qc.cx(1,0)
    qc.ry(-phi/2, 1)
    qc.ry(-phi/2, 0)
    qc.cx(1,0)
    qc.s_dg(0)
    qc.rz(-np.pi/2,1)
    qc.sx_dg(1)
    qc.rz(np.pi/2,1)
    qc.rz(-beta,0)

    return Operation(
        name="xxyy", num_qubits=2, num_clbits=0, params=[phi,beta], definition=qc
    )

def Barrier(num_qubits=1):
    res = Operation(num_qubits=num_qubits, name="barrier")

    res.permeability = {i: False for i in range(num_qubits)}
    res.is_qfree = True

    return res


def u3Gate(theta=0, phi=0, lam=0):
    return U3Gate(theta, phi, lam)


def U1Gate(phi=0):
    res = RZGate(phi)
    res.name = "u1"
    res.params = [phi]
    return res


def Reset():
    res = Operation("reset", 1)
    res.permeability = {0: False}
    return res


def SXGate():
    res = RXGate(np.pi / 2)
    res.name = "sx"
    res.params = []
    return res


def SXDGGate():
    res = RXGate(-np.pi / 2)
    res.name = "sx_dg"
    res.params = []
    return res


def SwapGate():
    from qrisp.circuit.quantum_circuit import QuantumCircuit

    temp_qc = QuantumCircuit(2)
    temp_qc.cx(0, 1)
    temp_qc.cx(1, 0)
    temp_qc.cx(0, 1)

    res = temp_qc.to_gate(name="swap")

    res.inverse = types.MethodType(lambda self: self.copy(), res)

    return res


def QubitAlloc():
    res = Operation("qb_alloc", 1)
    res.unitary = np.eye(2, dtype=np.complex64)
    res.qfree = True
    res.permeability[0] = False
    return res


def QubitDealloc():
    res = Operation("qb_dealloc", 1)
    res.unitary = np.eye(2, dtype=np.complex64)
    res.qfree = True
    res.permeability[0] = False
    return res


op_list = [
    XGate,
    YGate,
    ZGate,
    CXGate,
    CYGate,
    CZGate,
    MCXGate,
    PGate,
    CPGate,
    u3Gate,
    HGate,
    RXGate,
    RYGate,
    RZGate,
    MCRXGate,
    SGate,
    TGate,
    RXXGate,
    RZZGate,
    SXGate,
    SXDGGate,
    XXYYGate,
    Barrier,
    Measurement,
    Reset,
    QubitAlloc,
    QubitDealloc,
    GPhaseGate,
    SwapGate,
    U1Gate,
    IDGate,
]



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import h, QuantumFloat, multi_measurement
import numpy as np


def test_comparison():
    def test_comparison_helper(qf_0, qf_1, comp):
        qf_0 = qf_0.duplicate()
        h(qf_0)

        if isinstance(qf_1, QuantumFloat):
            qf_1 = qf_1.duplicate()
            h(qf_1)

        if comp == "eq":
            qf_res = qf_0 == qf_1
        elif comp == "neq":
            qf_res = qf_0 != qf_1
        elif comp == "leq":
            qf_res = qf_0 <= qf_1
        elif comp == "geq":
            qf_res = qf_0 >= qf_1
        elif comp == "lt":
            qf_res = qf_0 < qf_1
        elif comp == "gt":
            qf_res = qf_0 > qf_1

        if isinstance(qf_1, QuantumFloat):
            mes_res = multi_measurement([qf_0, qf_1, qf_res])
            for a, b, c in mes_res.keys():
                print(a, b, c)
                if comp == "eq":
                    assert (a == b) == c
                elif comp == "neq":
                    assert (a != b) == c
                elif comp == "leq":
                    assert (a <= b) == c
                elif comp == "geq":
                    assert (a >= b) == c
                elif comp == "lt":
                    assert (a < b) == c
                elif comp == "gt":
                    assert (a > b) == c

            # Test correct phase behavior7
            statevector = qf_res.qs.statevector("array")
            angles = np.angle(
                statevector[
                    np.abs(statevector) > 1 / 2 ** ((qf_0.size + qf_1.size) / 2)
                ]
            )
            assert np.sum(np.abs(angles)) < 0.1
        else:
            mes_res = multi_measurement([qf_0, qf_res])
            for a, c in mes_res.keys():
                b = qf_1
                print(a, b, c)
                if comp == "eq":
                    assert (a == b) == c
                elif comp == "neq":
                    assert (a != b) == c
                elif comp == "leq":
                    assert (a <= b) == c
                elif comp == "geq":
                    assert (a >= b) == c
                elif comp == "lt":
                    assert (a < b) == c
                elif comp == "gt":
                    assert (a > b) == c

    a = QuantumFloat(5, 2, signed=False)
    b = QuantumFloat(5, -1, signed=True)
    # Test all the operators
    for comp in ["eq", "neq", "lt", "gt", "leq", "geq"]:
        test_comparison_helper(a, b, comp)

    # Test specific constellations of QuantumFloat parameters
    a = QuantumFloat(3, -1, signed=True)
    b = QuantumFloat(5, 1, signed=True)
    test_comparison_helper(a, b, "lt")

    a = QuantumFloat(5, 2, signed=False)
    b = QuantumFloat(5, -1, signed=True)
    test_comparison_helper(a, b, "lt")

    a = QuantumFloat(4, -4, signed=False)
    b = QuantumFloat(4, 0, signed=True)
    test_comparison_helper(a, b, "lt")

    a = QuantumFloat(4, 1, signed=True)
    b = QuantumFloat(3, -1, signed=False)
    test_comparison_helper(a, b, "lt")

    # Test semi-classical comparisons
    a = QuantumFloat(4, signed=False)
    b = 4
    test_comparison_helper(a, b, "lt")
    test_comparison_helper(a, b, "gt")

    a = QuantumFloat(4, signed=True)
    b = 4
    test_comparison_helper(a, b, "lt")
    test_comparison_helper(a, b, "gt")

    a = QuantumFloat(4, -2, signed=False)
    b = 4
    test_comparison_helper(a, b, "lt")
    test_comparison_helper(a, b, "gt")

    a = QuantumFloat(8, 3, signed=False)
    b = 16
    test_comparison_helper(a, b, "lt")
    test_comparison_helper(a, b, "gt")



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import logging
import threading

from thrift.server import TServer


class StoppableThriftServer(TServer.TThreadedServer):
    def __init__(*args, **kwargs):
        args[0].is_running = False
        TServer.TThreadedServer.__init__(*args, **kwargs)

    def serve(self):
        self.serverTransport.listen()

        self.stop_flag = False
        thread_list = []
        while True:
            if self.stop_flag:
                for t in thread_list:
                    t.join()
                break
            try:
                self.is_running = True
                client = self.serverTransport.accept()
                t = threading.Thread(target=self.handle, args=(client,))
                t.setDaemon(self.daemon)
                t.start()
                thread_list.append(t)
            except OSError:
                pass
            except KeyboardInterrupt:
                raise
            except Exception as x:
                logging.exception(x)

    def stop(self):
        self.stop_flag = True
        try:
            self.serverTransport.handle.shutdown(2)
        except OSError:
            pass
        self.serverTransport.handle.close()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import sympy as sp


# Transforms a given symbol into a qb object
# For this we require that there is some quantum variable
# in qv_list which has the same name as the symbol
# ie. if the symbol has the name g3
# qv_list needs to contain a quantum variable with name g
# then qubit with index 3 is returned
def symbol_to_qbit(symbol, qv_list):
    if isinstance(symbol, sp.core.power.Pow):
        symbol = symbol.args[0]

    var_name = symbol.name.split("%")

    for qv in qv_list:
        if var_name[0] == qv.name:
            return qv.reg[int(var_name[1])]

    raise Exception("Could not find matching qubit for symbol " + symbol.name)


def monomial_to_list(monom):
    if not isinstance(
        monom,
        (
            sp.core.mul.Mul,
            sp.core.power.Pow,
            sp.core.numbers.Number,
            sp.core.symbol.Symbol,
        ),
    ):
        raise Exception("Tried to convert invalid monomial to list")

    from sympy import preorder_traversal

    result_list = []
    while True:
        for x in preorder_traversal(monom):
            if isinstance(x, sp.core.symbol.Symbol):
                monom = monom / x
                result_list.append(x)
                break
        else:
            break

    return [monom] + result_list


# Retrieves a list of the symbols appearing in the
# given sympy expression in alphabetical order
def get_ordered_symbol_list(expr):
    symbol_list = list(expr.free_symbols)
    symbol_list.sort(key=lambda x: x.name)
    return symbol_list


# Filters powers out of polynomial
def filter_pow(expr):
    pow_dic = {}
    for sub_expr in sp.preorder_traversal(expr):
        if isinstance(sub_expr, sp.core.power.Pow):
            pow_dic.update({sub_expr: get_ordered_symbol_list(sub_expr)[0]})
    return expr.subs(pow_dic)


# Turns a polynomial expression into a list of lists where
# example: x0*x1*x3 + x2*x4 --> [[x0,x1,x3],[x2,x4]]
def expr_to_list(expr):
    if not check_for_polynomial(expr):
        raise Exception("Tried to turn a non-polynomial expression into a list")

    result_list = []

    expr = expr.expand()
    if isinstance(expr, sp.core.add.Add):
        # Go through each summand
        for e1 in expr.args:
            result_list.append(monomial_to_list(e1))
    else:
        result_list.append(monomial_to_list(expr))

    return result_list


# Check if the given expression is a polynomial
def check_for_polynomial(expr):
    for x in sp.preorder_traversal(expr):
        if not isinstance(
            x,
            (
                sp.core.mul.Mul,
                sp.core.add.Add,
                sp.core.symbol.Symbol,
                sp.core.numbers.Number,
                sp.core.power.Pow,
            ),
        ):
            return False
    return True
